# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml


class Optimizer(pgml.YamlBaseModel):
    """Optimizer configuration"""

    type: str = "sgd"
    """Optimizer type"""

    learning_rate: float = 0.01
    """Learning rate"""



import abc
import importlib
import json
import pathlib
from dataclasses import dataclass

import protogen
from yapf.yapflib.yapf_api import FormatCode

from py_gen_ml.logging.setup_logger import setup_logger
from py_gen_ml.plugin.common import field_requires_typing_import
from py_gen_ml.yaml.object_path import InsertAnyOfWithObjectPath

logger = setup_logger(__name__)


@dataclass
class GenTask:
    """
    Data class representing a generation task.

    Attributes:
        obj_path (str): The path to the object module.
        obj_name (str): The name of the object.
        path (str): The path to write the JSON schema.
    """
    obj_path: str
    obj_name: str
    path: str

    def generate(self) -> None:
        """
        Generate the JSON schema for the object.

        This method imports the object module, retrieves the object, and writes the JSON schema to the specified path.
        """
        obj_module = importlib.import_module(self.obj_path)
        obj = getattr(obj_module, self.obj_name)
        pathlib.Path(self.path).parent.mkdir(parents=True, exist_ok=True)
        pathlib.Path(self.path).write_text(
            json.dumps(
                obj.model_json_schema(schema_generator=InsertAnyOfWithObjectPath,),
                indent=2,
            ),
        )


class Generator(abc.ABC):
    """
    Abstract base class for all generators.

    This class defines the interface for all generators used in the plugin.

    Args:
        gen (protogen.Plugin): The protogen plugin instance.
    """

    def __init__(self, gen: protogen.Plugin) -> None:
        """
        Initialize the Generator.

        Args:
            gen (protogen.Plugin): The protogen plugin instance.
        """
        self._gen = gen
        self._json_schema_gen_tasks: list[GenTask] = []
        self._source_root = self._gen.parameter['source_root']
        self._output_dir = self._gen.parameter['output_dir']
        self._configs_dir = self._gen.parameter['configs_dir']

    def generate_code(self) -> None:
        """
        Generate code for all relevant files.

        This method iterates through the files to generate, checks if the
        extension is enabled, and then generates the corresponding code.
        """
        for file in self._gen.files_to_generate:
            self._generate_code_for_file(file)

    def _requires_typing_import(self, file: protogen.File) -> bool:
        for message in file.messages:
            for field in message.fields:
                if field_requires_typing_import(field):
                    return True
        return False

    @abc.abstractmethod
    def _generate_code_for_file(self, file: protogen.File) -> None:
        """
        Generate code for a specific file.

        Args:
            file (protogen.File): The file to generate code for.
        """

    def _prepend_python_import(self, obj_path: str) -> str:
        prefix = str(pathlib.Path(self._output_dir).relative_to(self._source_root)).replace('/', '.')
        return f'{prefix}.{obj_path}'

    def _add_json_schema_gen_task(self, obj_path: str, obj_name: str, path: str) -> None:
        self._json_schema_gen_tasks.append(GenTask(self._prepend_python_import(obj_path), obj_name, path))

    @property
    def json_schema_gen_tasks(self) -> list[GenTask]:
        """
        Get the list of JSON schema generation tasks.

        Returns:
            List[GenTask]: The list of JSON schema generation tasks.
        """
        return self._json_schema_gen_tasks

    def _run_yapf(self, file: protogen.GeneratedFile) -> None:
        """
        Run yapf on the given file.

        Args:
            file (protogen.File): The file to run yapf on.
        """
        lines = file._buf
        yapf_toml_path = pathlib.Path(__file__).parent / 'yapf_style.toml'
        formatted_content, changed = FormatCode('\n'.join(lines), style_config=str(yapf_toml_path.absolute()))
        if changed:
            file._buf = formatted_content.split('\n')  # type: ignore


class InitGenerator(Generator):
    """
    Generator for the __init__.py file.

    This class is responsible for generating the __init__.py file,
    which is used to initialize the package.
    """

    def generate_code(self) -> None:
        """
        Generate the __init__.py file.

        This method creates a new generated file for the __init__.py file.
        It imports necessary modules and defines the __init__ module.
        """
        self._gen.new_generated_file('__init__.py', protogen.PyImportPath('__init__'))

    def _generate_code_for_file(self, file: protogen.File) -> None:
        pass



from py_gen_ml.plugin.generator import GenTask

json_schema_gen_tasks = [
    GenTask(obj_path="pgml_out.advanced_base", obj_name="LinearBlock", path="docs/snippets/configs/base/schemas/linear_block.json"),
    GenTask(obj_path="pgml_out.advanced_base", obj_name="Optimizer", path="docs/snippets/configs/base/schemas/optimizer.json"),
    GenTask(obj_path="pgml_out.advanced_base", obj_name="MLP", path="docs/snippets/configs/base/schemas/mlp.json"),
    GenTask(obj_path="pgml_out.advanced_base", obj_name="Training", path="docs/snippets/configs/base/schemas/training.json"),
    GenTask(obj_path="pgml_out.advanced_patch", obj_name="LinearBlockPatch", path="docs/snippets/configs/patch/schemas/linear_block.json"),
    GenTask(obj_path="pgml_out.advanced_patch", obj_name="OptimizerPatch", path="docs/snippets/configs/patch/schemas/optimizer.json"),
    GenTask(obj_path="pgml_out.advanced_patch", obj_name="MLPPatch", path="docs/snippets/configs/patch/schemas/mlp.json"),
    GenTask(obj_path="pgml_out.advanced_patch", obj_name="TrainingPatch", path="docs/snippets/configs/patch/schemas/training.json"),
    GenTask(obj_path="pgml_out.advanced_sweep", obj_name="LinearBlockSweep", path="docs/snippets/configs/sweep/schemas/linear_block.json"),
    GenTask(obj_path="pgml_out.advanced_sweep", obj_name="OptimizerSweep", path="docs/snippets/configs/sweep/schemas/optimizer.json"),
    GenTask(obj_path="pgml_out.advanced_sweep", obj_name="MLPSweep", path="docs/snippets/configs/sweep/schemas/mlp.json"),
    GenTask(obj_path="pgml_out.advanced_sweep", obj_name="TrainingSweep", path="docs/snippets/configs/sweep/schemas/training.json"),
]


# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import example_project.modules


class LinearBlock(pgml.YamlBaseModel):
    """Linear block configuration"""

    in_features: int
    """Number of input features"""

    out_features: int
    """Number of output features"""

    bias: bool
    """Bias"""

    dropout: float
    """Dropout probability"""

    activation: str
    """Activation function"""

    def build(self) -> "example_project.modules.LinearBlock":
        import example_project.modules

        return example_project.modules.LinearBlock(
            in_features=self.in_features,
            out_features=self.out_features,
            bias=self.bias,
            dropout=self.dropout,
            activation=self.activation,
        )


class MLP(pgml.YamlBaseModel):
    """MLP configuration"""

    layers: typing.List[LinearBlock]
    """Linear blocks"""



# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml


class MLPQuickstart(pgml.YamlBaseModel):
    """Multi-layer perceptron configuration"""

    num_layers: int
    """Number of layers"""

    num_units: int
    """Number of units"""

    activation: str
    """Activation function"""



import importlib
import pathlib
import sys
from typing import List, Optional, Set

import grpc_tools
import grpc_tools.protoc
import typer

import py_gen_ml as pgml
from py_gen_ml.logging.setup_logger import setup_logger

logger = setup_logger(__name__)

app = typer.Typer(pretty_exceptions_enable=False)


@app.command()
def generate(
    proto_file: List[str] = typer.Argument(..., help='Path to the protobuf file.'),
    proto_root: Optional[str] = typer.Option(
        None,
        help=(
            'Path to the root of the protobuf files which will be passed to the'
            'protoc command as an include path. If not specified, the script '
            'will try to infer it from the proto_file arguments by adding the '
            'parent of the proto_file arguments.'
        ),
    ),
    code_dir: str = typer.Option(
        'src/pgml_out',
        help='Path to the generated code directory.',
    ),
    source_root: str = typer.Option(
        'src',
        help='Path to the root of the source code.',
    ),
    configs_dir: str = typer.Option(
        'configs',
        help='Path to the base directory for configs.',
    ),
    export_code_generator_request: bool = typer.Option(
        False,
        help='Export the code generator request to a file.',
        hidden=True,
    ),
    export_code_generator_request_file: str = typer.Option(
        'code_generator_request.pbbin',
        help='Path to the file to export the code generator request to.',
        hidden=True,
    ),
) -> None:
    """
    Generate Pydantic models from a protobuf definition.

    This command is the core of the `py-gen-ml` toolbox. It is used to generate Pydantic models
    from a protobuf definition. By default, it generates models in the `src/pgml_out`
    directory. If your proto is called `example.proto`, it generates the following files:

    - `src/pgml_out/example_base.py` for a base model that follows the protobuf definition
    - `src/pgml_out/example_sweep.py` for a sweep model that can be used to sweep the base model
    - `src/pgml_out/example_patch.py` for a patch model that can be used to patch the base model

    If you've set the CLI option on a message called `Foo`, it will also generate

    - `src/pgml_out/example_cli_args.py` for CLI argument models
    - `src/pgml_out/foo_entrypoint.py` for an entrypoint that combines the base config, sweep, and CLI arguments.

    Other than that, it will generate JSON schemas in the `configs` directory.

    The structure of the schemas is as follows:

    - `configs/base/schemas/<message_name>.json`
    - `configs/patch/schemas/<message_name>.json`
    - `configs/sweep/schemas/<message_name>.json`
    """
    if not pathlib.Path(code_dir).resolve().is_relative_to(pathlib.Path(source_root).resolve()):
        raise ValueError(
            'The code_dir option must be a subdirectory of the source_root option.',
        )

    # Expand any glob patterns in the proto_file arguments
    proto_files: list[pathlib.Path] = []
    for path in proto_file:
        as_path = pathlib.Path(path)
        if as_path.is_absolute():
            root = pathlib.Path('/')
        else:
            root = pathlib.Path.cwd()
        proto_files.extend(root.glob(str(as_path)))

    if len(proto_files) == 0:
        raise ValueError('No protobuf files found.')

    if proto_root is None:
        proto_file_parents: Set[str] = set()
        for pf in proto_files:
            proto_file_parents.add(str(pf.parent))
        pgml_root = pathlib.Path(pgml.__file__).parent

        parent_cli_options = [f'-I{parent}' for parent in proto_file_parents]
    else:
        parent_cli_options = [f'-I{proto_root}']

    pgml_root = pathlib.Path(pgml.__file__).parent
    export_cmd_args = [] if not export_code_generator_request else [
        f'--py-ml_opt=export_code_generator_request={export_code_generator_request}',
        f'--py-ml_opt=export_code_generator_request_file={export_code_generator_request_file}',
    ]
    cmd = [
        f"-I{pathlib.Path(grpc_tools.__file__).parent / '_proto'}",
        f"-I{grpc_tools.protoc._get_resource_file_name('grpc_tools', '_proto')}",
        *parent_cli_options,
        f'-I{str(pgml_root.parent.absolute())}',
        '--plugin=py-ml',
        f'--py-ml_out={code_dir}',
        f'--py-ml_opt=output_dir={code_dir}',
        f'--py-ml_opt=source_root={source_root}',
        f'--py-ml_opt=configs_dir={configs_dir}',
        '--experimental_allow_proto3_optional',
        *export_cmd_args,
        *map(str, proto_files),
    ]
    pathlib.Path(code_dir).mkdir(parents=True, exist_ok=True)

    sys.path.append(str(pathlib.Path(source_root).absolute()))
    logger.debug(f"Running protoc with {' '.join(cmd)}")
    grpc_tools.protoc.main(cmd)
    python_import_path = str(
        pathlib.Path(code_dir).relative_to(pathlib.Path(source_root)),
    ).replace('/', '.') + '.__json_schema_tasks__'
    sys.path.append(source_root)
    module = importlib.import_module(python_import_path)
    tasks = module.json_schema_gen_tasks
    for task in tasks:
        task.generate()


if __name__ == '__main__':
    app()



from __future__ import annotations

import re
from dataclasses import dataclass, field, replace
from typing import Any, Dict, Final, Optional, TypeVar

from pydantic import BaseModel
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaMode
from pydantic_core.core_schema import CoreSchema

INTERNAL_OBJECT_PATH_PATTERN: Final[
    str] = r'^#\/?(?:[\$a-zA-Z0-9_]+(?:\[\d+\])?|\.\.)(?:\/([a-zA-Z0-9_]+(?:\[\d+\])?|\.\.))*$'  # noqa: E501
INTERNAL_OBJECT_PATH_REGEX: Final = re.compile(INTERNAL_OBJECT_PATH_PATTERN)
EXTERNAL_OBJECT_PATH_PATTERN: Final[
    str
] = r'^\!(\/?(?:[\w\.]+|\.\.)+(?:/(?:[\w\.]+|\.\.))*)' + f'({INTERNAL_OBJECT_PATH_PATTERN[1:-1]})?$'  # noqa: E501
EXTERNAL_OBJECT_PATH_REGEX: Final = re.compile(EXTERNAL_OBJECT_PATH_PATTERN)

TBaseModel = TypeVar('TBaseModel', bound=BaseModel)


@dataclass
class ObjectStep:
    """A step in an object path."""

    name: str
    """The name of the step."""

    index: Optional[int] = None
    """The index of the step."""


@dataclass
class ObjectPath:
    """A path to an object."""

    steps: list[ObjectStep] = field(default_factory=list)
    """The steps to the object."""

    is_absolute: bool = False
    """Whether the path is absolute."""

    def add_step(self, name: str) -> ObjectPath:
        """
        Add a step to the object path.

        Args:
            name (str): The name of the step to add.

        Returns:
            ObjectPath: The object path with the added step.
        """
        return replace(self, steps=self.steps.copy() + [ObjectStep(name=name)])

    def add_index_step(self, index: int) -> ObjectPath:
        """
        Add an index step to the object path.

        Args:
            index (int): The index to add.

        Returns:
            ObjectPath: The object path with the added index step.
        """
        if len(self.steps) == 0:
            raise ValueError('Cannot add an index step to an empty path')
        steps = self.steps.copy()
        steps[-1].index = index
        return replace(self, steps=steps)

    @classmethod
    def from_string(cls, path: str) -> ObjectPath:
        """
        Create an object path from a string.

        Args:
            path (str): The path to create.

        Returns:
            ObjectPath: The created object path.
        """
        steps: list[ObjectStep] = []

        path = path.removeprefix('#')
        if path.startswith('/'):
            is_absolute = True
        else:
            is_absolute = False
        for part in path.removeprefix('/').split('/'):
            key = None

            int_key_regex = re.compile(r'^(\w+)\[(\d+)\]$')  # matches "key[0]"
            if (match := int_key_regex.match(part)) is not None:
                part, key = match.groups()
                key = int(key)
            steps.append(ObjectStep(name=part, index=key))
        return cls(steps=steps, is_absolute=is_absolute)

    def concat(self, other: ObjectPath) -> ObjectPath:
        """
        Concatenate two object paths.

        Args:
            other (ObjectPath): The other object path to concatenate.

        Returns:
            ObjectPath: The concatenated object path.
        """
        if other.is_absolute:
            return other
        return ObjectPath(steps=self.steps.copy() + other.steps.copy())


def resolve_inner_ref(root: Dict[str, Any], path_to_ref: ObjectPath, path_at_ref: ObjectPath) -> Any:
    """
    Resolve a reference in a nested structure that is rooted at `root`.

    Args:
        root (Dict[str, Any]): The root of the nested structure.
        path_to_ref (ObjectPath): The path to the reference.
        path_at_ref (ObjectPath): The path at the reference.

    Returns:
        Any: The resolved value.
    """
    path_to_ref = ObjectPath(path_to_ref.steps.copy()[:-1])  # remove the last step, as it is the reference to itself
    while len(path_at_ref.steps) > 0 and len(path_to_ref.steps) > 0 and path_at_ref.steps[0].name == '..':
        path_at_ref = ObjectPath(steps=path_at_ref.steps.copy()[1:])
        path_to_ref = ObjectPath(steps=path_to_ref.steps.copy()[:-1])

    path = path_to_ref.concat(path_at_ref)

    current = root
    for step in path.steps:
        try:
            current = current[step.name]
        except KeyError:
            raise ValueError(f'Key {step.name} not found in {current}')
        if step.index is not None:
            current = current[step.index]
    return current


class InsertAnyOfWithObjectPath(GenerateJsonSchema):
    """
    A class that inserts anyOf with object path schemas into a JSON schema.

    With the injected schema, one can add references to objects within a YAML file. The references
    can be internal and external.
    """

    def generate(
        self,
        schema: CoreSchema,
        mode: JsonSchemaMode = 'validation',
    ) -> Dict[str, Any]:
        """
        Generate a JSON schema with anyOf with object path schemas.

        Args:
            schema (Dict[str, Any]): The schema to generate the anyOf with object path schemas for.
            mode (str): The mode to generate the schema for.

        Returns:
            Dict[str, Any]: The generated schema.
        """
        json_schema = super().generate(schema, mode)
        self._traverse_properties(json_schema['properties'], mode)
        for definition in json_schema.get('$defs', {}).values():
            if 'properties' not in definition:
                continue
            self._traverse_properties(definition['properties'], mode)

        json_schema['properties']['_defs_'] = {}
        return json_schema

    def _get_external_object_path_schema(self) -> Dict[str, Any]:
        return {
            'type': 'string',
            'description': 'An object path to an external object',
            'minLength': 2,
            'pattern': EXTERNAL_OBJECT_PATH_PATTERN,
        }

    def _get_internal_object_path_schema(self) -> Dict[str, Any]:
        return {
            'type': 'string',
            'description': 'An object path to an internal object',
            'minLength': 2,
            'pattern': INTERNAL_OBJECT_PATH_PATTERN,
        }

    def _traverse_properties(self, properties: Dict[str, Any], mode: str) -> None:
        # Traverse the schema and insert the object path schema
        for property in properties.values():
            if 'anyOf' in property:
                property['anyOf'].append(self._get_external_object_path_schema())
                property['anyOf'].append(self._get_internal_object_path_schema())
            else:
                inner_values = property.copy()
                if inner_values.get('type') == 'array':
                    property['items'] = {
                        'anyOf': [
                            inner_values['items'].copy(),
                            self._get_external_object_path_schema(),
                            self._get_internal_object_path_schema(),
                        ],
                    }
                else:
                    property.clear()
                    property['anyOf'] = [
                        inner_values,
                        self._get_external_object_path_schema(),
                        self._get_internal_object_path_schema(),
                    ]



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class Transformer(pgml.YamlBaseModel):
    """Transformer configuration"""

    num_layers: int
    """Number of layers"""

    num_heads: int
    """Number of heads"""

    activation: str
    """Activation function"""


class ConvBlock(pgml.YamlBaseModel):
    """Conv block"""

    out_channels: int
    """Number of output channels"""

    kernel_size: int
    """Kernel size"""

    activation: str
    """Activation function"""


class ConvNet(pgml.YamlBaseModel):
    """Convolutional neural network configuration"""

    layers: typing.List[ConvBlock]
    """Conv layer configuration"""


class Model(pgml.YamlBaseModel):
    """Model configuration"""

    backbone: typing.Union[Transformer, ConvNet]



from pathlib import Path
import shutil
from typing import List
import typer 


import mkdocs_gen_files


nav = mkdocs_gen_files.Nav()

def main(
    include: List[str] = ["**/*.py"],
    exclude: List[str] = [
        "py_gen_ml/cmd/*.py",
        "py_gen_ml/plugin/*.py",
        "py_gen_ml/typing/*.py",
        "py_gen_ml/logging/*.py",
        "**/*_pb2.py",
    ],
) -> None:
    root = Path(__file__).parents[2]
    source_dir = root / "src"
    included_files = set()
    for include_pattern in include:
        included_files.update(Path(source_dir).glob(include_pattern))
    
    excluded_files = set()
    for exclude_pattern in exclude:
        excluded_files.update(Path(source_dir).glob(exclude_pattern))
    
    files = included_files - excluded_files

    source_dir = Path(source_dir)
    for path in files: 
        module_path = path.relative_to(source_dir).with_suffix("")
        doc_path = path.relative_to(source_dir).with_suffix(".md")
        full_doc_path = Path("reference", doc_path)
        
        parts = tuple(module_path.parts)

        if parts[-1] == "__init__":
            parts = parts[:-1]
            doc_path = doc_path.with_name("index.md")
            full_doc_path = full_doc_path.with_name("index.md")
        elif parts[-1] == "__main__":
            continue

        nav[parts] = doc_path.as_posix()        
        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            ident = ".".join(parts)
            fd.write(f"::: {ident}")
            
        mkdocs_gen_files.set_edit_path(full_doc_path, path.relative_to(root))

        
    with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
        nav_file.writelines(nav.build_literate_nav())

main()


from py_gen_ml.plugin.generator import GenTask

json_schema_gen_tasks = [
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Int32Test', path='configs/base/schemas/int32_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Int64Test', path='configs/base/schemas/int64_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Uint32Test', path='configs/base/schemas/uint32_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Uint64Test', path='configs/base/schemas/uint64_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Sint32Test', path='configs/base/schemas/sint32_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Sint64Test', path='configs/base/schemas/sint64_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Fixed32Test', path='configs/base/schemas/fixed32_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='Fixed64Test', path='configs/base/schemas/fixed64_test.json'),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Sfixed32Test',
        path='configs/base/schemas/sfixed32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Sfixed64Test',
        path='configs/base/schemas/sfixed64_test.json',
    ),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='BoolTest', path='configs/base/schemas/bool_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='FloatTest', path='configs/base/schemas/float_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='DoubleTest', path='configs/base/schemas/double_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='StringTest', path='configs/base/schemas/string_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='BytesTest', path='configs/base/schemas/bytes_test.json'),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='EnumTest', path='configs/base/schemas/enum_test.json'),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='EnumDefaultTest',
        path='configs/base/schemas/enum_default_test.json',
    ),
    GenTask(obj_path='pgml_out_test.unit_base', obj_name='OneofTest', path='configs/base/schemas/oneof_test.json'),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='RepeatedTest',
        path='configs/base/schemas/repeated_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='OptionalTest',
        path='configs/base/schemas/optional_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Int32DefaultTest',
        path='configs/base/schemas/int32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Int64DefaultTest',
        path='configs/base/schemas/int64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Uint32DefaultTest',
        path='configs/base/schemas/uint32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Uint64DefaultTest',
        path='configs/base/schemas/uint64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Sint32DefaultTest',
        path='configs/base/schemas/sint32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Sint64DefaultTest',
        path='configs/base/schemas/sint64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Fixed32DefaultTest',
        path='configs/base/schemas/fixed32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Fixed64DefaultTest',
        path='configs/base/schemas/fixed64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Sfixed32DefaultTest',
        path='configs/base/schemas/sfixed32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='Sfixed64DefaultTest',
        path='configs/base/schemas/sfixed64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='BoolDefaultTest',
        path='configs/base/schemas/bool_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='FloatDefaultTest',
        path='configs/base/schemas/float_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='DoubleDefaultTest',
        path='configs/base/schemas/double_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='StringDefaultTest',
        path='configs/base/schemas/string_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='BytesDefaultTest',
        path='configs/base/schemas/bytes_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='OneofDefaultTest',
        path='configs/base/schemas/oneof_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='OptionalDefaultTest',
        path='configs/base/schemas/optional_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='ExplicitCLIArgTest',
        path='configs/base/schemas/explicit_cli_arg_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='ImplicitCLIArgTest',
        path='configs/base/schemas/implicit_cli_arg_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='NestedFooTest',
        path='configs/base/schemas/nested_foo_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='NestedBarTest',
        path='configs/base/schemas/nested_bar_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_base',
        obj_name='RepeatedNestedBarTest',
        path='configs/base/schemas/repeated_nested_bar_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Int32TestPatch',
        path='configs/patch/schemas/int32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Int64TestPatch',
        path='configs/patch/schemas/int64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Uint32TestPatch',
        path='configs/patch/schemas/uint32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Uint64TestPatch',
        path='configs/patch/schemas/uint64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sint32TestPatch',
        path='configs/patch/schemas/sint32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sint64TestPatch',
        path='configs/patch/schemas/sint64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Fixed32TestPatch',
        path='configs/patch/schemas/fixed32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Fixed64TestPatch',
        path='configs/patch/schemas/fixed64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sfixed32TestPatch',
        path='configs/patch/schemas/sfixed32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sfixed64TestPatch',
        path='configs/patch/schemas/sfixed64_test.json',
    ),
    GenTask(obj_path='pgml_out_test.unit_patch', obj_name='BoolTestPatch', path='configs/patch/schemas/bool_test.json'),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='FloatTestPatch',
        path='configs/patch/schemas/float_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='DoubleTestPatch',
        path='configs/patch/schemas/double_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='StringTestPatch',
        path='configs/patch/schemas/string_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='BytesTestPatch',
        path='configs/patch/schemas/bytes_test.json',
    ),
    GenTask(obj_path='pgml_out_test.unit_patch', obj_name='EnumTestPatch', path='configs/patch/schemas/enum_test.json'),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='EnumDefaultTestPatch',
        path='configs/patch/schemas/enum_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='OneofTestPatch',
        path='configs/patch/schemas/oneof_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='RepeatedTestPatch',
        path='configs/patch/schemas/repeated_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='OptionalTestPatch',
        path='configs/patch/schemas/optional_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Int32DefaultTestPatch',
        path='configs/patch/schemas/int32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Int64DefaultTestPatch',
        path='configs/patch/schemas/int64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Uint32DefaultTestPatch',
        path='configs/patch/schemas/uint32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Uint64DefaultTestPatch',
        path='configs/patch/schemas/uint64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sint32DefaultTestPatch',
        path='configs/patch/schemas/sint32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sint64DefaultTestPatch',
        path='configs/patch/schemas/sint64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Fixed32DefaultTestPatch',
        path='configs/patch/schemas/fixed32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Fixed64DefaultTestPatch',
        path='configs/patch/schemas/fixed64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sfixed32DefaultTestPatch',
        path='configs/patch/schemas/sfixed32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='Sfixed64DefaultTestPatch',
        path='configs/patch/schemas/sfixed64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='BoolDefaultTestPatch',
        path='configs/patch/schemas/bool_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='FloatDefaultTestPatch',
        path='configs/patch/schemas/float_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='DoubleDefaultTestPatch',
        path='configs/patch/schemas/double_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='StringDefaultTestPatch',
        path='configs/patch/schemas/string_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='BytesDefaultTestPatch',
        path='configs/patch/schemas/bytes_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='OneofDefaultTestPatch',
        path='configs/patch/schemas/oneof_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='OptionalDefaultTestPatch',
        path='configs/patch/schemas/optional_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='ExplicitCLIArgTestPatch',
        path='configs/patch/schemas/explicit_cli_arg_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='ImplicitCLIArgTestPatch',
        path='configs/patch/schemas/implicit_cli_arg_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='NestedFooTestPatch',
        path='configs/patch/schemas/nested_foo_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='NestedBarTestPatch',
        path='configs/patch/schemas/nested_bar_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_patch',
        obj_name='RepeatedNestedBarTestPatch',
        path='configs/patch/schemas/repeated_nested_bar_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Int32TestSweep',
        path='configs/sweep/schemas/int32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Int64TestSweep',
        path='configs/sweep/schemas/int64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Uint32TestSweep',
        path='configs/sweep/schemas/uint32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Uint64TestSweep',
        path='configs/sweep/schemas/uint64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sint32TestSweep',
        path='configs/sweep/schemas/sint32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sint64TestSweep',
        path='configs/sweep/schemas/sint64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Fixed32TestSweep',
        path='configs/sweep/schemas/fixed32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Fixed64TestSweep',
        path='configs/sweep/schemas/fixed64_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sfixed32TestSweep',
        path='configs/sweep/schemas/sfixed32_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sfixed64TestSweep',
        path='configs/sweep/schemas/sfixed64_test.json',
    ),
    GenTask(obj_path='pgml_out_test.unit_sweep', obj_name='BoolTestSweep', path='configs/sweep/schemas/bool_test.json'),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='FloatTestSweep',
        path='configs/sweep/schemas/float_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='DoubleTestSweep',
        path='configs/sweep/schemas/double_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='StringTestSweep',
        path='configs/sweep/schemas/string_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='BytesTestSweep',
        path='configs/sweep/schemas/bytes_test.json',
    ),
    GenTask(obj_path='pgml_out_test.unit_sweep', obj_name='EnumTestSweep', path='configs/sweep/schemas/enum_test.json'),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='EnumDefaultTestSweep',
        path='configs/sweep/schemas/enum_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='OneofTestSweep',
        path='configs/sweep/schemas/oneof_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='RepeatedTestSweep',
        path='configs/sweep/schemas/repeated_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='OptionalTestSweep',
        path='configs/sweep/schemas/optional_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Int32DefaultTestSweep',
        path='configs/sweep/schemas/int32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Int64DefaultTestSweep',
        path='configs/sweep/schemas/int64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Uint32DefaultTestSweep',
        path='configs/sweep/schemas/uint32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Uint64DefaultTestSweep',
        path='configs/sweep/schemas/uint64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sint32DefaultTestSweep',
        path='configs/sweep/schemas/sint32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sint64DefaultTestSweep',
        path='configs/sweep/schemas/sint64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Fixed32DefaultTestSweep',
        path='configs/sweep/schemas/fixed32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Fixed64DefaultTestSweep',
        path='configs/sweep/schemas/fixed64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sfixed32DefaultTestSweep',
        path='configs/sweep/schemas/sfixed32_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='Sfixed64DefaultTestSweep',
        path='configs/sweep/schemas/sfixed64_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='BoolDefaultTestSweep',
        path='configs/sweep/schemas/bool_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='FloatDefaultTestSweep',
        path='configs/sweep/schemas/float_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='DoubleDefaultTestSweep',
        path='configs/sweep/schemas/double_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='StringDefaultTestSweep',
        path='configs/sweep/schemas/string_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='BytesDefaultTestSweep',
        path='configs/sweep/schemas/bytes_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='OneofDefaultTestSweep',
        path='configs/sweep/schemas/oneof_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='OptionalDefaultTestSweep',
        path='configs/sweep/schemas/optional_default_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='ExplicitCLIArgTestSweep',
        path='configs/sweep/schemas/explicit_cli_arg_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='ImplicitCLIArgTestSweep',
        path='configs/sweep/schemas/implicit_cli_arg_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='NestedFooTestSweep',
        path='configs/sweep/schemas/nested_foo_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='NestedBarTestSweep',
        path='configs/sweep/schemas/nested_bar_test.json',
    ),
    GenTask(
        obj_path='pgml_out_test.unit_sweep',
        obj_name='RepeatedNestedBarTestSweep',
        path='configs/sweep/schemas/repeated_nested_bar_test.json',
    ),
]



# Autogenerated code. DO NOT EDIT.
import typing

import py_gen_ml as pgml

from . import unit_base as base


class Int32TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Int64TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Uint32TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Uint64TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sint32TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sint64TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Fixed32TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Fixed64TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sfixed32TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sfixed64TestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class BoolTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[bool] = None


class FloatTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[float] = None


class DoubleTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[float] = None


class StringTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[str] = None


class BytesTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[bytes] = None


class EnumTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[base.Enum] = None


class EnumDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[base.Enum] = None


class OneofTestPatch(pgml.YamlBaseModel):
    value: typing.Union[int, str]


class RepeatedTestPatch(pgml.YamlBaseModel):
    values: typing.Optional[list[int]] = None


class OptionalTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Int32DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Int64DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Uint32DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Uint64DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sint32DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sint64DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Fixed32DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Fixed64DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sfixed32DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Sfixed64DefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class BoolDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[bool] = None


class FloatDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[float] = None


class DoubleDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[float] = None


class StringDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[str] = None


class BytesDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[bytes] = None


class OneofDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Union[int, str]


class OptionalDefaultTestPatch(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class ExplicitCLIArgTestPatch(pgml.YamlBaseModel):
    bar: typing.Optional[str] = None


class ImplicitCLIArgTestPatch(pgml.YamlBaseModel):
    bar: typing.Optional[str] = None


class NestedFooTestPatch(pgml.YamlBaseModel):
    foo: typing.Optional[str] = None


class NestedBarTestPatch(pgml.YamlBaseModel):
    foo_0: typing.Optional[NestedFooTestPatch] = None
    foo_1: typing.Optional[NestedFooTestPatch] = None


class RepeatedNestedBarTestPatch(pgml.YamlBaseModel):
    bar: typing.Optional[list[NestedBarTestPatch]] = None



import os
import pathlib
import typing
import uuid
from typing import Any, List, Optional

import optuna
import pgml_out.config_base as base
import pgml_out.config_cli_args as cli_args
import pgml_out.config_sweep as sweep
import rich
import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data
import torchmetrics
import torchmetrics.classification
import tqdm
import typer
from cifar10.data import get_data_loader, get_transform
from cifar10.modules import CLASSES, Model

import py_gen_ml as pgml


class Trainer:

    def __init__(
        self,
        model: torch.nn.Module,
        train_loader: torch.utils.data.DataLoader[Any],
        test_loader: torch.utils.data.DataLoader[Any],
        criterion: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        accuracy_metric_train: torchmetrics.classification.MulticlassAccuracy,
        accuracy_metric_test: torchmetrics.classification.MulticlassAccuracy,
        train_loss_metric: torchmetrics.MeanMetric,
        test_loss_metric: torchmetrics.MeanMetric,
        trial: Optional[optuna.Trial] = None,
    ) -> None:
        self._train_loader = train_loader
        self._test_loader = test_loader
        self._model = model
        self._criterion = criterion
        self._optimizer = optimizer
        self._accuracy_metric_train = accuracy_metric_train
        self._accuracy_metric_test = accuracy_metric_test
        self._train_loss_metric = train_loss_metric
        self._test_loss_metric = test_loss_metric
        self._trial = trial

    def train(self, num_epochs: int) -> float:
        device = get_device()
        step = 0
        for epoch in tqdm.trange(num_epochs, position=0, desc='Epoch'):
            batch_bar = tqdm.tqdm(self._train_loader, position=1, desc='Batch')
            for inputs, labels in batch_bar:
                inputs = inputs.to(device)
                labels = labels.to(device)

                self._optimizer.zero_grad()
                outputs = self._model(inputs)
                loss = self._criterion(outputs, labels)
                loss.backward()
                self._optimizer.step()

                self._train_loss_metric.update(loss)
                self._accuracy_metric_train.update(outputs, labels)
                step += 1

                if step % 10 == 0:
                    batch_bar.set_postfix(
                        loss=self._train_loss_metric.compute().item(),
                        accuracy=self._accuracy_metric_train.compute().item(),
                    )
                    self._train_loss_metric.reset()
                    self._accuracy_metric_train.reset()

            self._evaluate()
            if self._trial is not None:
                self._trial.report(self._accuracy_metric_test.compute().item(), epoch)
        return self._accuracy_metric_test.compute().item()

    @torch.inference_mode()
    def _evaluate(self) -> None:
        self._model.eval()
        self._accuracy_metric_test.reset()
        self._test_loss_metric.reset()
        for images, labels in tqdm.tqdm(self._test_loader, position=1, desc='Evaluating'):
            images = images.to(get_device())
            labels = labels.to(get_device())
            outputs = self._model(images)
            loss = self._criterion(outputs, labels)
            self._test_loss_metric.update(loss)
            self._accuracy_metric_test.update(outputs, labels)
        print(f'Test accuracy: {self._accuracy_metric_test.compute().item()}')
        print(f'Test loss: {self._test_loss_metric.compute().item()}')
        self._model.train()


def get_accuracy_metric(num_classes: int) -> torchmetrics.Metric:
    device = get_device()
    return torchmetrics.classification.MulticlassAccuracy(num_classes=num_classes).to(device)


def train_model(project: base.Project, trial: typing.Optional[optuna.Trial] = None) -> float:
    rich.print(project)

    transform = get_transform()
    train_loader = get_data_loader(transform=transform, batch_size=project.data.batch_size, train=True)
    test_loader = get_data_loader(transform=transform, batch_size=project.data.batch_size, train=False)

    device = get_device()
    print(f'device {device}')

    model = Model.from_config(config=project.net).to(device)
    path = pathlib.Path(f"{os.environ['HOME']}/gen_ml/logs/{uuid.uuid4()}")
    print(f'Storing logs at {path}')

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(params=model.parameters(), lr=project.optimizer.learning_rate)
    accuracy_metric_train = get_accuracy_metric(num_classes=len(CLASSES))
    accuracy_metric_test = get_accuracy_metric(num_classes=len(CLASSES))
    train_loss_metric = torchmetrics.MeanMetric().to(device)
    test_loss_metric = torchmetrics.MeanMetric().to(device)
    trainer = Trainer(
        model=model,
        train_loader=train_loader,
        test_loader=test_loader,
        criterion=criterion,
        optimizer=optimizer,
        accuracy_metric_train=accuracy_metric_train,
        accuracy_metric_test=accuracy_metric_test,
        train_loss_metric=train_loss_metric,
        test_loss_metric=test_loss_metric,
        trial=trial,
    )
    accuracy = trainer.train(num_epochs=project.data.num_epochs)
    return accuracy


app = typer.Typer(pretty_exceptions_enable=False)


@pgml.pgml_cmd(app=app)
def main(
    config_paths: List[str] = typer.Option(..., help='Paths to config files'),
    sweep_paths: List[str] = typer.Option(default_factory=list, help='Paths to sweep files'),
    cli_args: cli_args.ProjectArgs = typer.Option(...),
) -> None:
    config = base.Project.from_yaml_files(config_paths)
    config = config.apply_cli_args(cli_args)

    if len(sweep_paths) == 0:
        train_model(config, trial=None)
        return

    sweep_config = sweep.ProjectSweep.from_yaml_files(sweep_paths)

    def objective(trial: optuna.Trial) -> float:
        sampler = pgml.OptunaSampler(trial=trial)
        patch = sampler.sample(sweep_config)
        accuracy = train_model(project=config.merge(patch), trial=trial)
        return accuracy

    study = optuna.create_study(direction='maximize')
    study.optimize(objective, n_trials=100)


def get_device() -> torch.device:
    return torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')


if __name__ == '__main__':
    app()



# Autogenerated code. DO NOT EDIT.
import enum

import py_gen_ml as pgml

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import cifar10.modules


class Activation(str, enum.Enum):
    """Activation function"""

    GELU = "GELU"
    """GELU activation"""

    RELU = "RELU"
    """ReLU activation"""


class Optimizer(pgml.YamlBaseModel):
    """Optimizer configuration"""

    learning_rate: float = 0.0001
    """Learning rate"""

    beta1: float = 0.99
    """Decay rate"""


class Data(pgml.YamlBaseModel):
    """Data configuration"""

    batch_size: int = 32
    """Batch size for a single GPU"""

    num_epochs: int = 10
    """Number of epochs to train"""


class ConvBlock(pgml.YamlBaseModel):
    """Convolutional layer configuration"""

    out_channels: int = 128
    """Number of output channels"""

    kernel_size: int = 3
    """Square kernel size"""

    pool_size: int = 2
    """Square pool size"""

    activation: Activation = Activation.GELU
    """Activation function"""

    def build(self) -> "cifar10.modules.ConvBlock":
        import cifar10.modules

        return cifar10.modules.ConvBlock(
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            pool_size=self.pool_size,
            activation=self.activation,
        )


class LinearBlock(pgml.YamlBaseModel):
    """Linear layer configuration"""

    out_features: int = 128
    """Number of output features"""

    activation: Activation = Activation.GELU
    """Activation function"""

    def build(self) -> "cifar10.modules.LinearBlock":
        import cifar10.modules

        return cifar10.modules.LinearBlock(
            out_features=self.out_features,
            activation=self.activation,
        )


class ConvNet(pgml.YamlBaseModel):
    """Convolutional neural network configuration"""

    block: ConvBlock
    """Conv layer configuration"""

    num_layers: int = 2
    """Number of layers"""


class MLP(pgml.YamlBaseModel):
    """Multi-layer perceptron configuration"""

    block: LinearBlock
    """Linear layer configuration"""

    num_layers: int = 2
    """Number of layers"""


class Model(pgml.YamlBaseModel):
    """Model configuration"""

    conv_net: ConvNet
    """Conv blocks"""

    head: MLP
    """MLP head"""


class Project(pgml.YamlBaseModel):
    """Global configuration"""

    net: Model
    """Model configuration"""

    optimizer: Optimizer
    """Optimizer configuration"""

    data: Data
    """Data configuration"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class TransformerPatch(pgml.YamlBaseModel):
    """Transformer configuration"""

    num_layers: typing.Optional[int] = None
    """Number of layers"""

    num_heads: typing.Optional[int] = None
    """Number of heads"""

    activation: typing.Optional[str] = None
    """Activation function"""


class ConvBlockPatch(pgml.YamlBaseModel):
    """Conv block"""

    out_channels: typing.Optional[int] = None
    """Number of output channels"""

    kernel_size: typing.Optional[int] = None
    """Kernel size"""

    activation: typing.Optional[str] = None
    """Activation function"""


class ConvNetPatch(pgml.YamlBaseModel):
    """Convolutional neural network configuration"""

    layers: typing.Optional[typing.List[ConvBlockPatch]] = None
    """Conv layer configuration"""


class ModelPatch(pgml.YamlBaseModel):
    """Model configuration"""

    backbone: typing.Union[TransformerPatch, ConvNetPatch]



# Autogenerated code. DO NOT EDIT.
import enum

import py_gen_ml as pgml


class Activation(str, enum.Enum):
    """Activation function"""

    RELU = "RELU"
    """Rectified Linear Unit"""

    GELU = "GELU"
    """Gaussian Error Linear Unit"""


class Linear(pgml.YamlBaseModel):
    """Linear layer"""

    in_features: int
    """Number of input features"""

    out_features: int
    """Number of output features"""

    activation: Activation = Activation.GELU
    """Activation function"""



import protogen

from py_gen_ml.plugin.generator import Generator, GenTask


class JsonSchemaTaskGenerator(Generator):
    """
    A generator that creates tasks for generating JSON schema from a set of objects.

    The generated file will be imported by the py-gen-ml CLI to generate JSON schema for the given objects.
    """

    def __init__(self, gen: protogen.Plugin, tasks: list[GenTask]) -> None:
        """
        Initialize the JsonSchemaTaskGenerator.

        Args:
            gen (protogen.Plugin): The plugin instance.
            tasks (List[GenTask]): The list of tasks to generate.
        """
        super().__init__(gen)
        self._tasks = tasks

    def generate_code(self) -> None:
        """Generate the code for the JSON schema tasks."""
        g = self._gen.new_generated_file('__json_schema_tasks__.py', protogen.PyImportPath('__json_schema_tasks__'))
        g.P('from py_gen_ml.plugin.generator import GenTask')
        g.P()
        g.P('json_schema_gen_tasks = [')
        g.set_indent(4)
        for task in self._tasks:
            g.P(f"GenTask(obj_path=\"{task.obj_path}\", obj_name=\"{task.obj_name}\", path=\"{task.path}\"),")
        g.set_indent(0)
        g.P(']')

    def _generate_code_for_file(self, file: protogen.File) -> None:
        pass



# Autogenerated code. DO NOT EDIT.
import typing

import pydantic
import typer

import py_gen_ml as pgml


class ExplicitCLIArgTestArgs(pgml.YamlBaseModel):
    foo: typing.Annotated[
        typing.Optional[str],
        typer.Option(help='Maps to bar'),
        pydantic.Field(None),
        pgml.ArgRef('bar'),
    ]


class ImplicitCLIArgTestArgs(pgml.YamlBaseModel):
    bar: typing.Annotated[
        typing.Optional[str],
        typer.Option(help='Maps to bar'),
        pydantic.Field(None),
        pgml.ArgRef('bar'),
    ]



import textwrap
import typing
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Set, Tuple, TypeVar

import networkx
import protogen

from py_gen_ml.extensions_pb2 import CLI
from py_gen_ml.logging.setup_logger import setup_logger
from py_gen_ml.plugin.common import (
    WRAP_WIDTH,
    generate_docstring,
    get_element_subgraphs,
    get_extension_value,
    py_import_for_source_file_derived_file,
    snake_case,
)
from py_gen_ml.plugin.constants import (
    BASE_MODEL_ALIAS,
    CLI_ARGS_ALIAS,
    PGML_ALIAS,
    SWEEP_MODEL_ALIAS,
)
from py_gen_ml.plugin.generator import Generator
from py_gen_ml.typing.some import some

logger = setup_logger(__name__)

T = TypeVar('T')


@dataclass
class FieldPath:
    """A path to a field in a message."""
    path: List[str]
    field: protogen.Field


class CliArgsGenerator(Generator):
    """
    A generator that creates CLI argument models based on the input protobuf files.

    The generated file will be imported by the py-gen-ml CLI to generate CLI argument models for the given protobuf files.
    """

    def __init__(self, gen: protogen.Plugin, suffix: str) -> None:
        """
        Initialize the CliArgsGenerator.

        Args:
            gen (protogen.Plugin): The protogen plugin instance.
            suffix (str): The suffix to be added to the generated file names.

        This class is responsible for generating CLI argument models based on
        the input protobuf files. It creates pydantic models that can be used
        with typer for command-line interfaces.
        """
        super().__init__(gen)
        self._suffix = suffix

    def _generate_code_for_file(self, file: protogen.File) -> None:
        """
        Generate the args models corresponding to a file.

        Args:
            file (protogen.File): The input file.

        Returns:
            protogen.GeneratedFile: The generated file.
        """
        if not any(
            (cli := get_extension_value(message, 'cli', CLI)) is not None and cli.enable for message in file.messages
        ):
            return
        g = self._gen.new_generated_file(
            file.proto.name.replace('.proto', self._suffix),
            file.py_import_path,
        )
        g.P('# Autogenerated code. DO NOT EDIT.')
        g.P(f'import py_gen_ml as {PGML_ALIAS}')
        g.P('import typing')
        g.P()
        g.P('import pydantic')
        g.P('import typer')
        g.P()

        import_file_path = str(Path(file.proto.name.replace('.proto', '_base')).name)
        g.P(f'from . import {import_file_path} as {BASE_MODEL_ALIAS}')
        g.P()
        g.P()

        dependency_subgraphs = get_element_subgraphs(file)
        for subgraph in dependency_subgraphs:
            for message in networkx.topological_sort(subgraph):
                if not isinstance(message, protogen.Message):
                    continue
                self._generate_message_args(g, message, subgraph)
                self._generate_entrypoint_script_for_message(message)

        self._run_yapf(g)

    def _generate_message_args(
        self,
        g: protogen.GeneratedFile,
        message: protogen.Message,
        graph: networkx.MultiDiGraph,
    ) -> None:
        """
        Generate the 'Args' counterpart of a given message.

        Args:
            g (protogen.GeneratedFile): The generated file to which we append the generated Python code.
            message (protogen.Message): The proto message that we are currently processing.
            graph (networkx.DiGraph): The networkx graph that contains the message.
        """
        cli_proto = get_extension_value(message, 'cli', CLI)
        if cli_proto is None or not cli_proto.enable:
            return

        g.P(f'class {message.proto.name}Args({PGML_ALIAS}.YamlBaseModel):')
        g.set_indent(4)

        generate_docstring(g, message)

        explicit_paths: Set[Tuple[str, ...]] = set()
        for arg in cli_proto.arg:
            path = arg.path.split('.')
            field = self._get_field_by_path(message, path)
            g.P(f'{arg.name}: {self.field_to_annotation(field, path)}')
            generate_docstring(g, field)
            explicit_paths.add(tuple(path))

        descendant_fields = self._gather_descendant_fields(message, graph, explicit_paths)
        for name, field_paths in descendant_fields.items():
            if len(field_paths) != 1:
                continue
            field_path = field_paths[0]
            g.P(f'{name}: {self.field_to_annotation(field=field_path.field, path=field_path.path)}')
            generate_docstring(g, field_path.field)

        g.P()
        g.P()
        g.set_indent(0)

    def _generate_entrypoint_script_for_message(self, message: protogen.Message) -> None:
        """
        Generate the entrypoint script for a given message.

        Args:
            message (protogen.Message): The message to generate the entrypoint script for.
        """
        cli_proto = get_extension_value(message, 'cli', CLI)
        if cli_proto is None or not cli_proto.enable:
            return
        g = self._gen.new_generated_file(
            f'{snake_case(message.proto.name)}_entrypoint.py',
            message.parent_file.py_import_path,
        )
        sweep_py_import = py_import_for_source_file_derived_file(g, '_sweep')
        sweep_py_import = self._prepend_python_import(sweep_py_import)

        base_py_import = py_import_for_source_file_derived_file(g, '_base')
        base_py_import = self._prepend_python_import(base_py_import)

        cli_args_py_import = py_import_for_source_file_derived_file(g, '_cli_args')
        cli_args_py_import = self._prepend_python_import(cli_args_py_import)

        g.P(f'import {base_py_import} as {BASE_MODEL_ALIAS}')
        g.P(f'import {sweep_py_import} as {SWEEP_MODEL_ALIAS}')
        g.P(f'import {cli_args_py_import} as {CLI_ARGS_ALIAS}')
        g.P(f'import typer')
        g.P(f'import py_gen_ml as {PGML_ALIAS}')
        g.P(f'import optuna')
        g.P('import typing')
        g.P()
        g.P(f'app = typer.Typer(pretty_exceptions_enable=False)')
        g.P()
        g.P(f'def run_trial(')
        g.set_indent(4)
        g.P(f'{snake_case(message.proto.name)}: {BASE_MODEL_ALIAS}.{message.proto.name},')
        g.P('trial: typing.Optional[optuna.Trial] = None')
        g.set_indent(0)
        g.P(') -> typing.Union[float, typing.Sequence[float]]:')
        g.set_indent(4)
        g.P(f"\"\"\"")
        lines = textwrap.wrap(
            f'Run a trial with the given values for {snake_case(message.proto.name)}. The sampled '
            f'hyperparameters have already been added to the trial.',
            width=WRAP_WIDTH - 4,
        )
        for line in lines:
            g.P(line)
        g.P("\"\"\"")
        g.P('# TODO: Implement this function')
        g.P('return 0.0')
        g.set_indent(0)
        g.P()
        g.P(f'@{PGML_ALIAS}.pgml_cmd(app=app)')
        g.P(f'def main(')
        g.set_indent(4)
        g.P(f'config_paths: typing.List[str] = typer.Option(..., help="Paths to config files"),')
        g.P(f'sweep_paths: typing.List[str] = typer.Option(')
        g.set_indent(8)
        g.P('default_factory=list,')
        g.P('help="Paths to sweep files"')
        g.set_indent(4)
        g.P('),')
        g.P(f'cli_args: {CLI_ARGS_ALIAS}.{message.proto.name}Args = typer.Option(...),')
        g.set_indent(0)
        g.P(f') -> None:')
        g.set_indent(4)
        g.P(f'{snake_case(message.proto.name)} = {BASE_MODEL_ALIAS}.{message.proto.name}.from_yaml_files(config_paths)')
        g.P(f'{snake_case(message.proto.name)} = {snake_case(message.proto.name)}.apply_cli_args(cli_args)')
        g.P(f'if len(sweep_paths) == 0:')
        g.set_indent(8)
        g.P(f'run_trial({snake_case(message.proto.name)})')
        g.P('return')
        g.set_indent(4)
        g.P(
            f'{snake_case(message.proto.name)}_sweep = {SWEEP_MODEL_ALIAS}.{message.proto.name}Sweep.from_yaml_files(sweep_paths)',
        )
        g.P()
        g.P('def objective(trial: optuna.Trial) -> typing.Union[')
        g.set_indent(8)
        g.P('float,')
        g.P('typing.Sequence[float]')
        g.set_indent(4)
        g.P(']:')
        g.set_indent(8)
        g.P(f'optuna_sampler = {PGML_ALIAS}.OptunaSampler(trial)')
        g.P(f'{snake_case(message.proto.name)}_patch = optuna_sampler.sample({snake_case(message.proto.name)}_sweep)')
        g.P(
            f'{snake_case(message.proto.name)}_patched = {snake_case(message.proto.name)}.merge({snake_case(message.proto.name)}_patch)',
        )
        g.P(f'objective_value = run_trial({snake_case(message.proto.name)}_patched, trial)')
        g.P('return objective_value')
        g.set_indent(4)
        g.P()
        g.P("study = optuna.create_study(direction='maximize')")
        g.P('study.optimize(objective, n_trials=100)')
        g.set_indent(0)
        g.P()
        g.P()
        g.P('if __name__ == "__main__":')
        g.set_indent(4)
        g.P('app()')

    def _gather_descendant_fields(
        self,
        message: protogen.Message,
        graph: networkx.MultiDiGraph,
        explicit_paths: Set[Tuple[str, ...]],
    ) -> Dict[str, list[FieldPath]]:
        """
        Gathers descendant fields of the given message.

        It goes down the dependency hierarchy to find field names at leaf nodes.

        Args:
            message (protogen.Message): The message of which to find descendant
                fields.
            graph (networkx.MultiDiGraph): A networkx graph that contains the message.

        Returns:
            OrderedDict[str, List[FieldPath]]: A mapping of field name to a list FieldPath objects.
        """
        fields = OrderedDict[str, List[FieldPath]]()
        for descendant in [message, *networkx.ancestors(graph, message)]:
            if isinstance(descendant, protogen.Enum):
                continue
            raw_paths = typing.cast(
                List[Tuple[Any, Any, str]],
                list(
                    networkx.all_simple_edge_paths(graph, source=descendant, target=message),
                ),
            )
            # Filter out paths that are already in explicit_paths
            paths = [tuple(edge[2] for edge in reversed(edges)) for edges in raw_paths]
            if len(paths) == 0:

                for field in descendant.fields:
                    if field.kind == protogen.Kind.MESSAGE:
                        continue

                    name = typing.cast(str, field.py_name)  # type: ignore
                    if (name,) in explicit_paths:
                        continue

                    if name in fields:
                        fields[name].append(FieldPath(path=[name], field=field))
                    else:
                        fields[name] = [FieldPath(path=[name], field=field)]
                continue
            elif len(paths) == 1:
                shortcut_paths = [()]
            else:
                max_len = max(len(path) for path in paths)
                for shortcut_len in range(1, max_len + 1):
                    shortcut_paths: list[tuple[str, ...]] = [edges[-shortcut_len:] for edges in paths]
                    if len(set(shortcut_paths)) == len(paths):
                        break
                else:
                    raise ValueError(f'Could not find a unique path for message {message.proto.name}')

            for field in descendant.fields:
                if field.kind == protogen.Kind.MESSAGE:
                    continue
                for shortcut_path, path in zip(shortcut_paths, paths):
                    field_py_name = typing.cast(str, field.py_name)  # type: ignore
                    if (*path, field_py_name) in explicit_paths:
                        continue
                    name = '_'.join([*shortcut_path, field_py_name])
                    if name in fields:
                        fields[name].append(FieldPath(path=[*path, field_py_name], field=field))
                    else:
                        fields[name] = [FieldPath(path=[*path, field_py_name], field=field)]
        return fields

    def field_to_annotation(self, field: protogen.Field, path: List[str]) -> str:
        """
        Get the annotation for a field and a path.

        The path is what the field maps to in the base model.

        Args:
            field (protogen.Field): Field in the protobuf Message
            path (List[str]): Path to the field in the base model that this field
                maps to.

        Returns:
            List[str]: A list of strings to append.
        """
        annotation = self.field_to_type(field)
        if field.is_list() and len(path) == 0:
            annotation = f'List[{annotation}]'
        typer_option = ''

        if len(path) > 0:
            path_str = '.'.join(path)
            if field.location.leading_comments == '':
                typer_option = f", typer.Option(help=\"Maps to {path_str}\")"
            else:
                typer_option = f", typer.Option(help=\"{field.location.leading_comments.rstrip('.')}. Maps to '{path_str}'\")"
            annotation = f"typing.Annotated[typing.Optional[{annotation}]{typer_option}, pydantic.Field(None), {PGML_ALIAS}.ArgRef(\"{path_str}\"),]"
        else:
            if field.location.leading_comments == '':
                typer_option = f", typer.Option(help=\"Maps to {field.py_name}\")"
            else:
                typer_option = f", typer.Option(help=\"{field.location.leading_comments.rstrip('.')}. Maps to '{field.py_name}'\")"
            annotation = f'typing.Annotated[typing.Optional[{annotation}]{typer_option}, pydantic.Field(None),]'
        return annotation

    def _get_field_by_path(self, message: protogen.Message, path: List[str]) -> protogen.Field:
        """
        Get the field at a given path in a message.

        Args:
            message (protogen.Message): The message to get the field from.
            path (List[str]): The path to the field.

        Returns:
            protogen.Field: The field at the given path.

        Raises:
            ValueError: The field at the given path was not found.
        """
        for field in message.fields:
            if field.py_name == path[0]:
                if len(path) == 1:
                    return field
                if field.kind == protogen.Kind.MESSAGE:
                    return self._get_field_by_path(some(field.message), path[1:])
        else:
            raise ValueError(f'Field {path[0]} not found in message {message.proto.name}')

    def field_to_type(self, field: protogen.Field) -> str:
        """
        Converts a protogen field to a type.

        Args:
            field (protogen.Field): Field in the proto message.

        Raises:
            ValueError: The field kind is not supported.

        Returns:
            str: A string representation of the type.
        """
        if field.kind == protogen.Kind.MESSAGE:
            return f'{some(field.message).py_ident.py_name}Args'
        elif field.kind == protogen.Kind.DOUBLE:
            return 'float'
        elif field.kind == protogen.Kind.FLOAT:
            return 'float'
        elif field.kind == protogen.Kind.INT64:
            return 'int'
        elif field.kind == protogen.Kind.UINT64:
            return 'int'
        elif field.kind == protogen.Kind.INT32:
            return 'int'
        elif field.kind == protogen.Kind.FIXED64:
            return 'int'
        elif field.kind == protogen.Kind.FIXED32:
            return 'int'
        elif field.kind == protogen.Kind.BOOL:
            return 'bool'
        elif field.kind == protogen.Kind.STRING:
            return 'str'
        elif field.kind == protogen.Kind.BYTES:
            return 'bytes'
        elif field.kind == protogen.Kind.UINT32:
            return 'int'
        elif field.kind == protogen.Kind.ENUM:
            return f'{BASE_MODEL_ALIAS}.{some(field.enum).py_ident.py_name}'
        elif field.kind == protogen.Kind.SFIXED32:
            return 'int'
        elif field.kind == protogen.Kind.SFIXED64:
            return 'int'
        elif field.kind == protogen.Kind.SINT32:
            return 'int'
        elif field.kind == protogen.Kind.SINT64:
            return 'int'
        else:
            raise ValueError(f'Unknown field kind: {field.kind}')



from typing import Self

import pgml_out.config_base as base
import torch

CLASSES = ['airplane', 'automobile', 'bird', 'cat', 'deer', 'dog', 'frog', 'horse', 'ship', 'truck']


def create_activation(activation: base.Activation) -> torch.nn.Module:
    if activation == base.Activation.GELU:
        return torch.nn.GELU()
    elif activation == base.Activation.RELU:
        return torch.nn.ReLU()
    else:
        raise ValueError(f'Invalid activation function: {activation}')


class LinearBlock(torch.nn.Module):
    """Linear layer with activation configuration"""

    def __init__(self, out_features: int, activation: base.Activation) -> None:
        super().__init__()
        self.linear = torch.nn.LazyLinear(out_features=out_features)
        self.activation = create_activation(activation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation(self.linear(x))


class ConvBlock(torch.nn.Module):
    """Convolutional layer with activation configuration"""

    def __init__(self, out_channels: int, kernel_size: int, pool_size: int, activation: base.Activation) -> None:
        super().__init__()
        self.conv = torch.nn.LazyConv2d(out_channels=out_channels, kernel_size=kernel_size)
        self.pool = torch.nn.MaxPool2d(kernel_size=pool_size)
        self.activation = create_activation(activation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation(self.pool(self.conv(x)))


class ConvNet(torch.nn.Module):

    def __init__(self, block: base.ConvBlock, num_layers: int) -> None:
        super().__init__()
        self.layers = [block.build() for _ in range(num_layers)]
        self.net = torch.nn.Sequential(*self.layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)


class MLP(torch.nn.Module):

    def __init__(self, block: base.LinearBlock, num_layers: int) -> None:
        super().__init__()
        self.layers = [block.build() for _ in range(num_layers)]
        self.net = torch.nn.Sequential(*self.layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)


class Model(torch.nn.Module):

    def __init__(self, conv_net: ConvNet, head: MLP) -> None:
        super().__init__()
        self.conv_net = conv_net
        self.head = head
        self.class_logits = torch.nn.LazyLinear(out_features=len(CLASSES),)
        self.flatten = torch.nn.Flatten()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv_net(x)
        x = self.head(x)
        x = self.flatten(x)
        x = self.class_logits(x)
        return x

    @classmethod
    def from_config(cls, config: base.Model) -> Self:
        conv_net = ConvNet(
            block=config.conv_net.block,
            num_layers=config.conv_net.num_layers,
        )
        head = MLP(
            block=config.head.block,
            num_layers=config.head.num_layers,
        )
        return cls(conv_net, head)



from __future__ import annotations

import abc
from dataclasses import dataclass
from dataclasses import replace as dataclass_replace
from typing import Any, Dict, Generic, List, Literal, Optional, TypeVar, Union

from pydantic import BaseModel

from py_gen_ml.yaml.yaml_model import YamlBaseModel

T = TypeVar('T')
Tb = TypeVar('Tb')
TScalar = TypeVar('TScalar', float, int, bool, str, bytes)
TBaseModel = TypeVar('TBaseModel', bound=BaseModel)


@dataclass
class SweepSamplerContext(Generic[T]):
    """A context for a sweep sampler."""

    path_parts: List[str]
    """The path parts of the sweep sampler context."""

    ctx: T
    """The context of the sweep sampler."""

    def step(self, name: str, index: Optional[int] = None) -> SweepSamplerContext[T]:
        """
        Step into a nested object.

        Args:
            name (str): The name of the object.
            index (Optional[int]): The index of the object.

        Returns:
            SweepSamplerContext: The context with the step.
        """
        if index is not None:
            name = f'{name}[{index}]'
        return dataclass_replace(self, path_parts=[*self.path_parts, name])

    @property
    def path(self) -> str:
        """
        Get the path to the current object.

        Returns:
            str: The path to the current object.
        """
        return '.'.join(self.path_parts)


class SweepModel(YamlBaseModel, Generic[T], abc.ABC):
    """
    A sweep model.

    This is a pydantic base model that defines sweeps for its fields.
    """

    @abc.abstractmethod
    def accept(self, visitor: SweepVisitor[Tb], context: SweepSamplerContext[Tb]) -> T:
        """
        Accept a visitor.

        Args:
            visitor (SweepVisitor[Tb]): The visitor to accept.
            context (SweepSamplerContext[Tb]): The context to use for the visitor.

        Returns:
            T: The result of the visitor.
        """


class SweepVisitor(abc.ABC, Generic[T]):
    """
    A sweep sampler.

    Defines a visitor pattern for sampling a model from a sweep configuration.
    """

    @abc.abstractmethod
    def visit_float_log_uniform(self, log_uniform: FloatLogUniform, context: SweepSamplerContext[T]) -> float:
        """
        Visit a float log uniform.

        Args:
            log_uniform (FloatLogUniform): The float log uniform to visit.
            context (SweepSamplerContext): The context to use for the visitor.

        Returns:
            float: The sampled float.
        """

    @abc.abstractmethod
    def visit_int_uniform(self, uniform: IntUniform, context: SweepSamplerContext[T]) -> int:
        """
        Visit an int uniform.

        Args:
            uniform (IntUniform): The int uniform to visit.
            context (SweepSamplerContext): The context to use for the visitor.

        Returns:
            int: The sampled int.
        """

    @abc.abstractmethod
    def visit_float_uniform(self, uniform: FloatUniform, context: SweepSamplerContext[T]) -> float:
        """
        Visit a float uniform.

        Args:
            uniform (FloatUniform): The float uniform to visit.
            context (SweepSamplerContext): The context to use for the visitor.

        Returns:
            float: The sampled float.
        """

    @abc.abstractmethod
    def visit_sweep_model(self, sweep_model: Sweeper[TBaseModel], context: SweepSamplerContext[T]) -> TBaseModel:
        """
        Visit a sweep model.

        Args:
            sweep_model (Sweeper[TBaseModel]): The sweep model to visit.
            context (SweepSamplerContext): The context to use for the visitor.

        Returns:
            TBaseModel: The sampled model.
        """

    @abc.abstractmethod
    def visit_nested_choice(
        self,
        sweep_choice: NestedChoice[TSweep, TBaseModel],
        context: SweepSamplerContext[T],
    ) -> TBaseModel:
        """
        Visit a nested choice.

        Args:
            sweep_choice (NestedChoice[TSweep, TBaseModel]): The nested choice to visit.
            context (SweepSamplerContext): The context to use for the visitor.

        Returns:
            TBaseModel: The sampled model.
        """

    @abc.abstractmethod
    def visit_choice(self, choice: Choice[TScalar], context: SweepSamplerContext[T]) -> TScalar:
        """
        Visit a choice.

        Args:
            choice (Choice[TScalar]): The choice to visit.
            context (SweepSamplerContext): The context to use for the visitor.

        Returns:
            TScalar: The sampled value.
        """


class SweepSampler(abc.ABC, Generic[TBaseModel]):
    """
    A sweep sampler.

    Defines a visitor pattern for sampling a model from a sweep configuration.
    """

    @abc.abstractmethod
    def sample(self, sweep_model: Sweeper[TBaseModel]) -> TBaseModel:
        """
        Sample a model.

        The sample should be a patch that is then applied to the corresponding base model.

        Args:
            sweep_model (Sweeper[TBaseModel]): The model to sample.

        Returns:
            TBaseModel: The sampled model.
        """


class Sweeper(SweepModel[TBaseModel]):
    """
    A sweeper.

    Sweepers allow for sampling a model from a set of options.
    """

    def accept(self, visitor: SweepVisitor[T], context: SweepSamplerContext[T]) -> TBaseModel:
        """
        Accept a visitor.

        Args:
            visitor (SweepVisitor[T]): The visitor to accept.
            context (SweepSamplerContext[T]): The context to use for the visitor.

        Returns:
            TBaseModel: The sampled model.
        """
        return visitor.visit_sweep_model(self, context)

    def new_base_model(self, **kwargs: Any) -> TBaseModel:
        """
        Create a new base model.

        Args:
            kwargs (Any): The keyword arguments to pass to the base model.

        Returns:
            TBaseModel: The new base model.
        """
        return self._get_base_type_from_generic_arg()(**kwargs)

    @classmethod
    def _get_base_type_from_generic_arg(cls) -> type[TBaseModel]:
        return cls.__bases__[0].__pydantic_generic_metadata__['args'][0]


class FloatLogUniform(SweepModel[float]):
    """
    A float log uniform.

    Float log uniforms allow for sampling a float value within a logarithmic range.
    """

    log_low: float
    log_high: float

    def accept(self, visitor: SweepVisitor[T], context: SweepSamplerContext[T]) -> float:
        """
        Visit a float log uniform.

        Args:
            visitor (SweepVisitor[T]): The visitor to accept.
            context (SweepSamplerContext[T]): The context to use for the visitor.

        Returns:
            float: The sampled float.
        """
        return visitor.visit_float_log_uniform(self, context=context)


class IntUniform(SweepModel[int]):
    """
    An int uniform.

    Int uniforms allow for sampling an int value within a range.
    """

    low: int
    high: int
    step: int = 1

    def accept(self, visitor: SweepVisitor[T], context: SweepSamplerContext[T]) -> int:
        """
        Visit an int uniform.

        Args:
            visitor (SweepVisitor[T]): The visitor to accept.
            context (SweepSamplerContext[T]): The context to use for the visitor.

        Returns:
            int: The sampled int.
        """
        return visitor.visit_int_uniform(self, context=context)


class FloatUniform(SweepModel[float]):
    """
    A float uniform.

    Float uniforms allow for sampling a float value within a range.
    """

    low: float
    high: float
    step: Optional[float] = None

    def accept(self, visitor: SweepVisitor[T], context: SweepSamplerContext[T]) -> float:
        """
        Visit a float uniform.

        Args:
            visitor (SweepVisitor[T]): The visitor to accept.
            context (SweepSamplerContext[T]): The context to use for the visitor.

        Returns:
            float: The sampled float.
        """
        return visitor.visit_float_uniform(self, context=context)


# In fact we'd like to have a higher-kinded type here, but that's not possible in Python without resorting
# to external libraries. So we just use a generic type variable here.
TSweep = TypeVar('TSweep', bound=Sweeper[BaseModel])


class NestedChoice(SweepModel[TBaseModel], Generic[TSweep, TBaseModel]):
    """
    A nested choice.

    Nested choices allow for sampling from a set of options by name. The structures
    that are sampled again are also sweep models, so they can be nested arbitrarily
    deep.
    """

    nested_options: Dict[str, TSweep]

    def accept(self, visitor: SweepVisitor[T], context: SweepSamplerContext[T]) -> TBaseModel:
        """
        Visit a nested choice.

        Args:
            visitor (SweepVisitor[T]): The visitor to accept.
            context (SweepSamplerContext[T]): The context to use for the visitor.

        Returns:
            TBaseModel: The sampled model.
        """
        return visitor.visit_nested_choice(self, context=context)


class Choice(SweepModel[TScalar]):
    """
    A choice.

    Choices allow for sampling from a set of options.
    """

    options: list[TScalar]

    def accept(self, visitor: SweepVisitor[T], context: SweepSamplerContext[T]) -> TScalar:
        """
        Visit a choice.

        Args:
            visitor (SweepVisitor[T]): The visitor to accept.
            context (SweepSamplerContext[T]): The context to use for the visitor.

        Returns:
            TScalar: The sampled value.
        """
        return visitor.visit_choice(self, context=context)


IntSweep = Union[int, Choice[int], IntUniform]
FloatSweep = Union[float, Choice[float], FloatLogUniform, FloatUniform]
BoolSweep = Union[bool, Literal['any']]
StrSweep = Union[str, Choice[str]]
BytesSweep = Union[bytes, Choice[bytes]]



from pathlib import Path

from pydantic import BaseModel

from .pgml_out_test import unit_base


def assert_base_model_eq(a: BaseModel, b: BaseModel) -> None:
    assert a.model_dump() == b.model_dump()


def test_nested_yaml_parsing(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text('foo_0:\n  foo: test_0\nfoo_1:\n  foo: test_1')
    parsed = unit_base.NestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.NestedBarTest(
            foo_0=unit_base.NestedFooTest(foo='test_0'),
            foo_1=unit_base.NestedFooTest(foo='test_1'),
        ),
    )


def test_nested_yaml_parsing_with_internal_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("foo_0:\n  foo: test_0\nfoo_1:\n  foo: '#/foo_0/foo'")
    parsed = unit_base.NestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.NestedBarTest(
            foo_0=unit_base.NestedFooTest(foo='test_0'),
            foo_1=unit_base.NestedFooTest(foo='test_0'),
        ),
    )


def test_nested_yaml_parsing_with_relative_internal_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("foo_0:\n  foo: test_0\nfoo_1:\n  foo: '#../foo_0/foo'")
    parsed = unit_base.NestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.NestedBarTest(
            foo_0=unit_base.NestedFooTest(foo='test_0'),
            foo_1=unit_base.NestedFooTest(foo='test_0'),
        ),
    )


def test_nested_yaml_parsing_with_def_internal_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("foo_0:\n  foo: test_0\nfoo_1:\n  foo: '#/_defs_/foo'\n_defs_:\n  foo: test_1")
    parsed = unit_base.NestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.NestedBarTest(
            foo_0=unit_base.NestedFooTest(foo='test_0'),
            foo_1=unit_base.NestedFooTest(foo='test_1'),
        ),
    )


def test_nested_yaml_external_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("foo_0:\n  foo: test_0\nfoo_1:\n  foo: '!external.yaml#/foo_0/foo'")
    external_path = tmp_path.joinpath('external.yaml')
    external_path.write_text('foo_0:\n  foo: test_1')
    parsed = unit_base.NestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.NestedBarTest(
            foo_0=unit_base.NestedFooTest(foo='test_0'),
            foo_1=unit_base.NestedFooTest(foo='test_1'),
        ),
    )


def test_repeated_nested_yaml_parsing(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text(
        """
bar:
  - foo_0:
      foo: test_0
    foo_1:
      foo: test_1
  - foo_0:
      foo: test_2
    foo_1:
      foo: test_3
""",
    )
    parsed = unit_base.RepeatedNestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.RepeatedNestedBarTest(
            bar=[
                unit_base.NestedBarTest(
                    foo_0=unit_base.NestedFooTest(foo='test_0'),
                    foo_1=unit_base.NestedFooTest(foo='test_1'),
                ),
                unit_base.NestedBarTest(
                    foo_0=unit_base.NestedFooTest(foo='test_2'),
                    foo_1=unit_base.NestedFooTest(foo='test_3'),
                ),
            ],
        ),
    )


def test_repeated_nested_yaml_parsing_with_internal_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("""
bar:
  - foo_0:
      foo: test_0
    foo_1:
      foo: '#/bar[0]/foo_0/foo'
""")
    parsed = unit_base.RepeatedNestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.RepeatedNestedBarTest(
            bar=[
                unit_base.NestedBarTest(
                    foo_0=unit_base.NestedFooTest(foo='test_0'),
                    foo_1=unit_base.NestedFooTest(foo='test_0'),
                ),
            ],
        ),
    )


def test_repeated_nested_yaml_parsing_with_relative_internal_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("""
bar:
  - foo_0:
      foo: test_0
    foo_1:
      foo: '#../../bar[0]/foo_0/foo'
""")
    parsed = unit_base.RepeatedNestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.RepeatedNestedBarTest(
            bar=[
                unit_base.NestedBarTest(
                    foo_0=unit_base.NestedFooTest(foo='test_0'),
                    foo_1=unit_base.NestedFooTest(foo='test_0'),
                ),
            ],
        ),
    )


def test_repeated_nested_yaml_parsing_with_def_internal_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text(
        """
bar:
  - foo_0:
      foo: test_0
    foo_1:
      foo: '#/_defs_/foo'
_defs_:
  foo: test_1
""",
    )
    parsed = unit_base.RepeatedNestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.RepeatedNestedBarTest(
            bar=[
                unit_base.NestedBarTest(
                    foo_0=unit_base.NestedFooTest(foo='test_0'),
                    foo_1=unit_base.NestedFooTest(foo='test_1'),
                ),
            ],
        ),
    )


def test_repeated_nested_yaml_external_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("""
bar:
  - foo_0:
      foo: test_0
    foo_1:
      foo: '!external.yaml#/foo_0/foo'
""")
    external_path = tmp_path.joinpath('external.yaml')
    external_path.write_text('foo_0:\n  foo: test_1')
    parsed = unit_base.RepeatedNestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.RepeatedNestedBarTest(
            bar=[
                unit_base.NestedBarTest(
                    foo_0=unit_base.NestedFooTest(foo='test_0'),
                    foo_1=unit_base.NestedFooTest(foo='test_1'),
                ),
            ],
        ),
    )


def test_repeated_external_whole_file_references(tmp_path: Path) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text("""
bar:
  - foo_0:
      foo: test_0
    foo_1: '!external.yaml'
""")
    external_path = tmp_path.joinpath('external.yaml')
    external_path.write_text('foo: test_1')
    parsed = unit_base.RepeatedNestedBarTest.from_yaml_file(str(config_path))
    assert_base_model_eq(
        parsed,
        unit_base.RepeatedNestedBarTest(
            bar=[
                unit_base.NestedBarTest(
                    foo_0=unit_base.NestedFooTest(foo='test_0'),
                    foo_1=unit_base.NestedFooTest(foo='test_1'),
                ),
            ],
        ),
    )



from py_gen_ml.plugin.generator import GenTask

json_schema_gen_tasks = [
    GenTask(obj_path="pgml_out.config_base", obj_name="Optimizer", path="configs/base/schemas/optimizer.json"),
    GenTask(obj_path="pgml_out.config_base", obj_name="Data", path="configs/base/schemas/data.json"),
    GenTask(obj_path="pgml_out.config_base", obj_name="ConvBlock", path="configs/base/schemas/conv_block.json"),
    GenTask(obj_path="pgml_out.config_base", obj_name="LinearBlock", path="configs/base/schemas/linear_block.json"),
    GenTask(obj_path="pgml_out.config_base", obj_name="ConvNet", path="configs/base/schemas/conv_net.json"),
    GenTask(obj_path="pgml_out.config_base", obj_name="MLP", path="configs/base/schemas/mlp.json"),
    GenTask(obj_path="pgml_out.config_base", obj_name="Model", path="configs/base/schemas/model.json"),
    GenTask(obj_path="pgml_out.config_base", obj_name="Project", path="configs/base/schemas/project.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="OptimizerPatch", path="configs/patch/schemas/optimizer.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="DataPatch", path="configs/patch/schemas/data.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="ConvBlockPatch", path="configs/patch/schemas/conv_block.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="LinearBlockPatch", path="configs/patch/schemas/linear_block.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="ConvNetPatch", path="configs/patch/schemas/conv_net.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="MLPPatch", path="configs/patch/schemas/mlp.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="ModelPatch", path="configs/patch/schemas/model.json"),
    GenTask(obj_path="pgml_out.config_patch", obj_name="ProjectPatch", path="configs/patch/schemas/project.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="OptimizerSweep", path="configs/sweep/schemas/optimizer.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="DataSweep", path="configs/sweep/schemas/data.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="ConvBlockSweep", path="configs/sweep/schemas/conv_block.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="LinearBlockSweep", path="configs/sweep/schemas/linear_block.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="ConvNetSweep", path="configs/sweep/schemas/conv_net.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="MLPSweep", path="configs/sweep/schemas/mlp.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="ModelSweep", path="configs/sweep/schemas/model.json"),
    GenTask(obj_path="pgml_out.config_sweep", obj_name="ProjectSweep", path="configs/sweep/schemas/project.json"),
]


# Autogenerated code. DO NOT EDIT.
import enum

import py_gen_ml as pgml


class Activation(str, enum.Enum):
    """Activation is an enum of activation functions."""

    RELU = "RELU"
    """ReLU is the Rectified Linear Unit activation function."""

    TANH = "TANH"
    """TANH is the hyperbolic tangent activation function."""

    SIGMOID = "SIGMOID"
    """SIGMOID is the sigmoid activation function."""


class MLPParsingDemo(pgml.YamlBaseModel):
    """MLP is a simple multi-layer perceptron."""

    num_layers: int = 2
    """Number of layers in the MLP."""

    num_units: int
    """Number of units in each layer."""

    activation: Activation
    """Activation function to use."""



import typing

import py_gen_ml as pgml

from . import quickstart_b_patch as patch
from . import quickstart_b_base as base


class MLPQuickstartSweep(pgml.Sweeper[patch.MLPQuickstartPatch]):
    """Multi-layer perceptron configuration"""

    num_layers: typing.Optional[pgml.IntSweep] = None
    """Number of layers"""

    num_units: typing.Optional[pgml.IntSweep] = None
    """Number of units"""

    activation: typing.Optional[pgml.StrSweep] = None
    """Activation function"""



MLPQuickstartSweepField = typing.Union[
    MLPQuickstartSweep,
    pgml.NestedChoice[MLPQuickstartSweep, patch.MLPQuickstartPatch],  # type: ignore
]




# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class MLPQuickstartPatch(pgml.YamlBaseModel):
    """Multi-layer perceptron configuration"""

    num_layers: typing.Optional[int] = None
    """Number of layers"""

    num_units: typing.Optional[int] = None
    """Number of units"""

    activation: typing.Optional[str] = None
    """Activation function"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class LinearPatch(pgml.YamlBaseModel):
    """Linear layer configuration"""

    in_features: typing.Optional[int] = None
    """Number of input features"""

    out_features: typing.Optional[int] = None
    """Number of output features"""

    bias: typing.Optional[bool] = None
    """Bias"""


class MLPPatch(pgml.YamlBaseModel):
    """MLP configuration"""

    layers: typing.Optional[typing.List[LinearPatch]] = None
    """Linear layers"""



import inspect
from typing import Any, Callable, Dict, Optional, Union

import typer
import typer.core
import typer.models
from pydantic import BaseModel

from py_gen_ml.typing.some import some


def pgml_cmd(
    app: typer.Typer,
    name: Optional[str] = None,
    *,
    cls: Optional[type[typer.core.TyperCommand]] = None,
    context_settings: Optional[Dict[Any, Any]] = None,
    help: Optional[str] = None,
    epilog: Optional[str] = None,
    short_help: Optional[str] = None,
    options_metavar: str = '[OPTIONS]',
    add_help_option: bool = True,
    no_args_is_help: bool = False,
    hidden: bool = False,
    deprecated: bool = False,
    rich_help_panel: Union[str, None] = typer.models.Default(None),
) -> Callable[[Callable[..., Any]], None]:
    """
    Decorator to create a Typer command from a function.

    This decorator creates a Typer command from a given function. It expects
    at least one parameter that is a Pydantic model. This model is used to
    parse command-line arguments.

    Other parameters are passed to the function as-is.

    Args:
        app (typer.Typer): The Typer app to add the command to.
        name (Optional[str]): The name of the command.
        cls (Optional[type[typer.core.TyperCommand]]): The Typer command class to use.
        context_settings (Optional[Dict[Any, Any]]): Context settings for the command.
        help (Optional[str]): The help text for the command.
        epilog (Optional[str]): The epilog text for the command.
        short_help (Optional[str]): The short help text for the command.
        options_metavar (str): The metavar for the options.
        add_help_option (bool): Whether to add a help option to the command.
        no_args_is_help (bool): When no arguments are provided, whether to show help.
        hidden (bool): Whether to hide the command from help messages.
        deprecated (bool): Whether to mark the command as deprecated.
        rich_help_panel (Union[str, None]): The rich help panel for the command.
    """

    def decorator(func: Callable[..., Any]) -> None:
        _command_from_func(
            app,
            func,
            name=name,
            cls=cls,
            context_settings=context_settings,
            help=help,
            epilog=epilog,
            short_help=short_help,
            options_metavar=options_metavar,
            add_help_option=add_help_option,
            no_args_is_help=no_args_is_help,
            hidden=hidden,
            deprecated=deprecated,
            rich_help_panel=rich_help_panel,
        )

    return decorator


def _command_from_func(
    app: typer.Typer,
    func: Callable[..., Any],
    name: Optional[str] = None,
    *,
    cls: Optional[type[typer.core.TyperCommand]] = None,
    context_settings: Optional[Dict[Any, Any]] = None,
    help: Optional[str] = None,
    epilog: Optional[str] = None,
    short_help: Optional[str] = None,
    options_metavar: str = '[OPTIONS]',
    add_help_option: bool = True,
    no_args_is_help: bool = False,
    hidden: bool = False,
    deprecated: bool = False,
    rich_help_panel: Union[str, None] = typer.models.Default(None),
) -> None:
    """
    Create a Typer command from a function.

    This function creates a Typer command from a given function. It expects
    at least one parameter that is a Pydantic model. This model is used to
    parse command-line arguments.

    Other parameters are passed to the function as-is.

    The function is then wrapped to accept both the Pydantic model and other
    arguments, allowing for a flexible command-line interface.

    Args:
        app (typer.Typer): The Typer app to add the command to.
        func (Callable[..., Any]): The function to create a command from.
        name (Optional[str]): The name of the command.
        cls (Optional[type[typer.core.TyperCommand]]): The Typer command class to use.
        context_settings (Optional[Dict[Any, Any]]): Context settings for the command.
        help (Optional[str]): The help text for the command.
        epilog (Optional[str]): The epilog text for the command.
        short_help (Optional[str]): The short help text for the command.
        options_metavar (str): The metavar for the options.
        add_help_option (bool): Whether to add a help option to the command.
        no_args_is_help (bool): When no arguments are provided, whether to show help.
        hidden (bool): Whether to hide the command from help messages.
        deprecated (bool): Whether to mark the command as deprecated.
        rich_help_panel (Union[str, None]): The rich help panel for the command.
    """
    func_sig = inspect.signature(func)

    # Find the Pydantic model parameter
    model_param_name = None
    model_class = None
    for param in func_sig.parameters.values():
        if not inspect.isclass(param.annotation) or not issubclass(param.annotation, BaseModel):
            continue
        if model_class is not None:
            raise ValueError('pgml_cmd supports only one base model')
        model_class = param.annotation
        model_param_name = param.name

    if model_class is None:
        raise ValueError('No Pydantic model found in function parameters')

    # Create a new function with combined parameters
    def command_fn(**kwargs: Any) -> None:
        # Extract arguments for the Pydantic model
        model_args = {k: v for k, v in kwargs.items() if k in model_class.model_fields}
        # Create an instance of the Pydantic model
        model_instance = model_class(**model_args)

        # Extract arguments for the function
        non_model_kwargs = {k: v for k, v in kwargs.items() if k in func_sig.parameters if k != model_param_name}
        func(**{some(model_param_name): model_instance, **non_model_kwargs})

    # Combine the parameters
    func_params = {k: v for k, v in func_sig.parameters.items() if k != model_param_name}
    combined_params = {**func_params, **model_class.__signature__.parameters}
    command_fn.__signature__ = inspect.Signature(list(combined_params.values()))  # type: ignore

    # Register the combined function with Typer
    app.command(
        name=name,
        cls=cls,
        context_settings=context_settings,
        help=help,
        epilog=epilog,
        short_help=short_help,
        options_metavar=options_metavar,
        add_help_option=add_help_option,
        no_args_is_help=no_args_is_help,
        hidden=hidden,
        deprecated=deprecated,
        rich_help_panel=rich_help_panel,
    )(command_fn)



# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml


class Model(pgml.YamlBaseModel):
    """Model configuration"""

    num_layers: int
    """Number of layers"""


class Training(pgml.YamlBaseModel):
    """Training configuration"""

    num_epochs: int
    """Number of epochs"""


class Dataset(pgml.YamlBaseModel):
    """Dataset configuration"""

    path: str
    """Path to the dataset"""


class Data(pgml.YamlBaseModel):
    """Data config"""

    train_dataset: Dataset
    """Path to the dataset"""

    test_dataset: Dataset
    """Path to the dataset"""

    num_workers: int
    """Number of workers for loading the dataset"""


class CliExtensionDemo(pgml.YamlBaseModel):
    """Global configuration"""

    data: Data
    """Dataset configuration"""

    model: Model
    """Model configuration"""

    training: Training
    """Training configuration"""



from pathlib import Path

import optuna
import pytest

import py_gen_ml as pgml
from py_gen_ml.sweep.sweep import FloatUniform, IntUniform

from .pgml_out_test import unit_base, unit_sweep


@pytest.mark.parametrize(
    'class_name',
    [
        'Int32Test',
        'Int64Test',
        'Uint32Test',
        'Uint64Test',
        'FloatTest',
        'DoubleTest',
        'Fixed32Test',
        'Fixed64Test',
        'Sfixed32Test',
        'Sfixed64Test',
    ],
)
def test_numeric_sweep(class_name: str) -> None:
    config = getattr(unit_base, class_name)(value=10)
    if class_name in ['FloatTest', 'DoubleTest']:
        sweep = getattr(unit_sweep, f'{class_name}Sweep')(value=FloatUniform(low=1, high=5, step=1))
    else:
        sweep = getattr(unit_sweep, f'{class_name}Sweep')(value=IntUniform(low=1, high=5, step=1))

    def trial_fn(trial: optuna.Trial) -> float:
        patch = pgml.OptunaSampler(trial).sample(sweep)
        patched_config = config.merge(patch)
        return patched_config.value

    study = optuna.create_study(direction='maximize',)
    study.optimize(trial_fn, n_trials=50)

    for trial in study.get_trials():
        assert trial.value is not None
        assert trial.value <= 5
        assert trial.value > 0


def test_bool_any_sweep(tmp_path: Path) -> None:
    config_path = tmp_path / 'config.yaml'
    config_path.write_text('value: true')
    sweep_path = tmp_path / 'sweep.yaml'
    sweep_path.write_text('value: any')

    config = unit_base.BoolTest.from_yaml_file(str(config_path))
    sweep = unit_sweep.BoolTestSweep.from_yaml_file(str(sweep_path))

    def trial_fn(trial: optuna.Trial) -> float:
        patch = pgml.OptunaSampler(trial).sample(sweep)
        patched_config = config.merge(patch)
        return 1.0 if patched_config.value else 0.0

    study = optuna.create_study(direction='maximize',)
    study.optimize(trial_fn, n_trials=50)

    for trial in study.get_trials():
        assert trial.value is not None
        assert trial.value == 1.0 or trial.value == 0.0


@pytest.mark.parametrize('class_name', [
    'StringTest',
    'BytesTest',
])
def test_string_bytes_sweep(class_name: str, tmp_path: Path) -> None:
    config_path = tmp_path / 'config.yaml'
    config_path.write_text(f'value: "hello"')
    sweep_path = tmp_path / 'sweep.yaml'
    sweep_path.write_text(f'value:\n  options: ["foo", "bar"]')

    config = getattr(unit_base, class_name).from_yaml_file(str(config_path))
    sweep = getattr(unit_sweep, f'{class_name}Sweep').from_yaml_file(str(sweep_path))

    def trial_fn(trial: optuna.Trial) -> float:
        patch = pgml.OptunaSampler(trial).sample(sweep)
        patched_config = config.merge(patch)
        assert patched_config.value not in ['hello', b'hello']
        return 1.0 if patched_config.value in ['foo', b'foo'] else 0.0

    study = optuna.create_study(direction='maximize',)
    study.optimize(trial_fn, n_trials=50)

    for trial in study.get_trials():
        assert trial.value is not None
        assert trial.value == 1.0 or trial.value == 0.0



import typing
from typing import NamedTuple

import pydantic

TBaseModel = typing.TypeVar('TBaseModel', bound=pydantic.BaseModel)


class ArgRef(NamedTuple):
    """A reference to an argument."""

    path: str
    """The path to the argument."""


def apply_args(model: TBaseModel, arg_ref_model: pydantic.BaseModel) -> TBaseModel:
    """
    Apply arguments from arg_ref_model to model.

    Args:
        model (TBaseModel): The model to apply the arguments to.
        arg_ref_model (pydantic.BaseModel): The model containing the arguments to apply.

    Returns:
        TBaseModel: The model with the applied arguments.
    """
    out = model.model_copy()
    for name, field_info in arg_ref_model.model_fields.items():
        # Find shortcuts and replace the values in the given model
        if any(isinstance(arg_ref := m, ArgRef) for m in getattr(field_info, 'metadata', [])):
            current = out
            *path_to_container, field_in_container = arg_ref.path.split('.')
            for part in path_to_container:
                current = getattr(current, part)
            value = getattr(arg_ref_model, name)
            if value is None:
                continue
            setattr(current, field_in_container, value)
        elif name in model.model_fields and (value := getattr(arg_ref_model, name)) is not None:
            setattr(out, name, value)

    return out



import typing

import py_gen_ml as pgml

from . import advanced_patch as patch
from . import advanced_base as base


class LinearBlockSweep(pgml.Sweeper[patch.LinearBlockPatch]):
    """Linear block configuration"""

    num_units: typing.Optional[pgml.IntSweep] = None
    """Number of units"""

    activation: typing.Optional[pgml.StrSweep] = None
    """Activation function"""


LinearBlockSweepField = typing.Union[
    LinearBlockSweep,
    pgml.NestedChoice[LinearBlockSweep, patch.LinearBlockPatch],  # type: ignore
]


class OptimizerSweep(pgml.Sweeper[patch.OptimizerPatch]):
    """Optimizer configuration"""

    type: typing.Optional[pgml.StrSweep] = None
    """Type of optimizer"""

    learning_rate: typing.Optional[pgml.FloatSweep] = None
    """Learning rate"""


OptimizerSweepField = typing.Union[
    OptimizerSweep,
    pgml.NestedChoice[OptimizerSweep, patch.OptimizerPatch],  # type: ignore
]


class MLPSweep(pgml.Sweeper[patch.MLPPatch]):
    """Multi-layer perceptron configuration"""

    layers: typing.Optional[LinearBlockSweepField] = None
    """List of linear blocks"""


MLPSweepField = typing.Union[
    MLPSweep,
    pgml.NestedChoice[MLPSweep, patch.MLPPatch],  # type: ignore
]


class TrainingSweep(pgml.Sweeper[patch.TrainingPatch]):
    """Training configuration"""

    mlp: typing.Optional[MLPSweepField] = None
    """Multi-layer perceptron configuration"""

    optimizer: typing.Optional[OptimizerSweepField] = None
    """Optimizer configuration"""



TrainingSweepField = typing.Union[
    TrainingSweep,
    pgml.NestedChoice[TrainingSweep, patch.TrainingPatch],  # type: ignore
]




# Autogenerated code. DO NOT EDIT.
import enum
import typing

import py_gen_ml as pgml


class Enum(str, enum.Enum):
    VALUE_1 = 'VALUE_1'
    VALUE_2 = 'VALUE_2'
    VALUE_3 = 'VALUE_3'


class Int32Test(pgml.YamlBaseModel):
    value: int


class Int64Test(pgml.YamlBaseModel):
    value: int


class Uint32Test(pgml.YamlBaseModel):
    value: int


class Uint64Test(pgml.YamlBaseModel):
    value: int


class Sint32Test(pgml.YamlBaseModel):
    value: int


class Sint64Test(pgml.YamlBaseModel):
    value: int


class Fixed32Test(pgml.YamlBaseModel):
    value: int


class Fixed64Test(pgml.YamlBaseModel):
    value: int


class Sfixed32Test(pgml.YamlBaseModel):
    value: int


class Sfixed64Test(pgml.YamlBaseModel):
    value: int


class BoolTest(pgml.YamlBaseModel):
    value: bool


class FloatTest(pgml.YamlBaseModel):
    value: float


class DoubleTest(pgml.YamlBaseModel):
    value: float


class StringTest(pgml.YamlBaseModel):
    value: str


class BytesTest(pgml.YamlBaseModel):
    value: bytes


class EnumTest(pgml.YamlBaseModel):
    value: Enum


class EnumDefaultTest(pgml.YamlBaseModel):
    value: Enum = Enum.VALUE_1


class OneofTest(pgml.YamlBaseModel):
    value: typing.Union[int, str]


class RepeatedTest(pgml.YamlBaseModel):
    values: list[int]


class OptionalTest(pgml.YamlBaseModel):
    value: typing.Optional[int] = None


class Int32DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Int64DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Uint32DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Uint64DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Sint32DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Sint64DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Fixed32DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Fixed64DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Sfixed32DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class Sfixed64DefaultTest(pgml.YamlBaseModel):
    value: int = 1


class BoolDefaultTest(pgml.YamlBaseModel):
    value: bool = True


class FloatDefaultTest(pgml.YamlBaseModel):
    value: float = 1.0


class DoubleDefaultTest(pgml.YamlBaseModel):
    value: float = 1.0


class StringDefaultTest(pgml.YamlBaseModel):
    value: str = 'test'


class BytesDefaultTest(pgml.YamlBaseModel):
    value: bytes = b'test'


class OneofDefaultTest(pgml.YamlBaseModel):
    value: typing.Union[int, str]


class OptionalDefaultTest(pgml.YamlBaseModel):
    value: typing.Optional[int] = 1


class ExplicitCLIArgTest(pgml.YamlBaseModel):
    bar: str


class ImplicitCLIArgTest(pgml.YamlBaseModel):
    bar: str


class NestedFooTest(pgml.YamlBaseModel):
    foo: str


class NestedBarTest(pgml.YamlBaseModel):
    foo_0: NestedFooTest
    foo_1: NestedFooTest


class RepeatedNestedBarTest(pgml.YamlBaseModel):
    bar: list[NestedBarTest]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: extensions.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'extensions.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x65xtensions.proto\x12\x04pgml\x1a google/protobuf/descriptor.proto\"!\n\x03\x41rg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\"-\n\x03\x43LI\x12\x16\n\x03\x61rg\x18\x01 \x03(\x0b\x32\t.pgml.Arg\x12\x0e\n\x06\x65nable\x18\x02 \x01(\x08\"\xbb\x02\n\rFieldDefaults\x12\x0f\n\x05\x66loat\x18\x01 \x01(\x02H\x00\x12\x10\n\x06\x64ouble\x18\x02 \x01(\x01H\x00\x12\x0f\n\x05int32\x18\x03 \x01(\x05H\x00\x12\x0f\n\x05int64\x18\x04 \x01(\x03H\x00\x12\x10\n\x06uint32\x18\x05 \x01(\rH\x00\x12\x10\n\x06uint64\x18\x06 \x01(\x04H\x00\x12\x10\n\x06sint32\x18\x07 \x01(\x11H\x00\x12\x10\n\x06sint64\x18\x08 \x01(\x12H\x00\x12\x11\n\x07\x66ixed32\x18\t \x01(\x07H\x00\x12\x11\n\x07\x66ixed64\x18\n \x01(\x06H\x00\x12\x12\n\x08sfixed32\x18\x0b \x01(\x0fH\x00\x12\x12\n\x08sfixed64\x18\x0c \x01(\x10H\x00\x12\x0e\n\x04\x62ool\x18\r \x01(\x08H\x00\x12\x10\n\x06string\x18\x0e \x01(\tH\x00\x12\x0f\n\x05\x62ytes\x18\x0f \x01(\x0cH\x00\x12\x0e\n\x04\x65num\x18\x10 \x01(\tH\x00\x42\x06\n\x04typeJ\x04\x08\x12\x10\x15:0\n\x06\x65nable\x12\x1c.google.protobuf.FileOptions\x18\xc3\x08 \x01(\x08\x88\x01\x01:1\n\x07\x66\x61\x63tory\x12\x1f.google.protobuf.MessageOptions\x18\xc3\x08 \x01(\t:8\n\x03\x63li\x12\x1f.google.protobuf.MessageOptions\x18\xc4\x08 \x01(\x0b\x32\t.pgml.CLI:G\n\x07\x64\x65\x66\x61ult\x12\x1d.google.protobuf.FieldOptions\x18\x93\t \x01(\x0b\x32\x13.pgml.FieldDefaults\x88\x01\x01:2\n\nas_varargs\x12\x1d.google.protobuf.FieldOptions\x18\x94\t \x01(\x08\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'extensions_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ARG']._serialized_start=60
  _globals['_ARG']._serialized_end=93
  _globals['_CLI']._serialized_start=95
  _globals['_CLI']._serialized_end=140
  _globals['_FIELDDEFAULTS']._serialized_start=143
  _globals['_FIELDDEFAULTS']._serialized_end=458
# @@protoc_insertion_point(module_scope)



import typing

import py_gen_ml as pgml

from . import config_patch as patch
from . import config_base as base

ActivationSweepField = typing.Union[
    pgml.Choice[base.Activation],
    typing.Literal['any'],
    base.Activation,
]


class OptimizerSweep(pgml.Sweeper[patch.OptimizerPatch]):
    """Optimizer configuration"""

    learning_rate: pgml.FloatSweep | None = None
    """Learning rate"""

    beta1: pgml.FloatSweep | None = None
    """Decay rate"""


OptimizerSweepField = typing.Union[
    OptimizerSweep,
    pgml.NestedChoice[OptimizerSweep, patch.OptimizerPatch],  # type: ignore
]


class DataSweep(pgml.Sweeper[patch.DataPatch]):
    """Data configuration"""

    batch_size: pgml.IntSweep | None = None
    """Batch size for a single GPU"""

    num_epochs: pgml.IntSweep | None = None
    """Number of epochs to train"""


DataSweepField = typing.Union[
    DataSweep,
    pgml.NestedChoice[DataSweep, patch.DataPatch],  # type: ignore
]


class ConvBlockSweep(pgml.Sweeper[patch.ConvBlockPatch]):
    """Convolutional layer configuration"""

    out_channels: pgml.IntSweep | None = None
    """Number of output channels"""

    kernel_size: pgml.IntSweep | None = None
    """Square kernel size"""

    pool_size: pgml.IntSweep | None = None
    """Square pool size"""

    activation: ActivationSweepField | None = None
    """Activation function"""


ConvBlockSweepField = typing.Union[
    ConvBlockSweep,
    pgml.NestedChoice[ConvBlockSweep, patch.ConvBlockPatch],  # type: ignore
]


class LinearBlockSweep(pgml.Sweeper[patch.LinearBlockPatch]):
    """Linear layer configuration"""

    out_features: pgml.IntSweep | None = None
    """Number of output features"""

    activation: ActivationSweepField | None = None
    """Activation function"""


LinearBlockSweepField = typing.Union[
    LinearBlockSweep,
    pgml.NestedChoice[LinearBlockSweep, patch.LinearBlockPatch],  # type: ignore
]


class ConvNetSweep(pgml.Sweeper[patch.ConvNetPatch]):
    """Convolutional neural network configuration"""

    block: ConvBlockSweepField | None = None
    """Conv layer configuration"""

    num_layers: pgml.IntSweep | None = None
    """Number of layers"""


ConvNetSweepField = typing.Union[
    ConvNetSweep,
    pgml.NestedChoice[ConvNetSweep, patch.ConvNetPatch],  # type: ignore
]


class MLPSweep(pgml.Sweeper[patch.MLPPatch]):
    """Multi-layer perceptron configuration"""

    block: LinearBlockSweepField | None = None
    """Linear layer configuration"""

    num_layers: pgml.IntSweep | None = None
    """Number of layers"""


MLPSweepField = typing.Union[
    MLPSweep,
    pgml.NestedChoice[MLPSweep, patch.MLPPatch],  # type: ignore
]


class ModelSweep(pgml.Sweeper[patch.ModelPatch]):
    """Model configuration"""

    conv_net: ConvNetSweepField | None = None
    """Conv blocks"""

    head: MLPSweepField | None = None
    """MLP head"""


ModelSweepField = typing.Union[
    ModelSweep,
    pgml.NestedChoice[ModelSweep, patch.ModelPatch],  # type: ignore
]


class ProjectSweep(pgml.Sweeper[patch.ProjectPatch]):
    """Global configuration"""

    net: ModelSweepField | None = None
    """Model configuration"""

    optimizer: OptimizerSweepField | None = None
    """Optimizer configuration"""

    data: DataSweepField | None = None
    """Data configuration"""



ProjectSweepField = typing.Union[
    ProjectSweep,
    pgml.NestedChoice[ProjectSweep, patch.ProjectPatch],  # type: ignore
]




import typing

import py_gen_ml as pgml

from . import default_patch as patch
from . import default_base as base


class OptimizerSweep(pgml.Sweeper[patch.OptimizerPatch]):
    """Optimizer configuration"""

    type: typing.Optional[pgml.StrSweep] = None
    """Optimizer type"""

    learning_rate: typing.Optional[pgml.FloatSweep] = None
    """Learning rate"""



OptimizerSweepField = typing.Union[
    OptimizerSweep,
    pgml.NestedChoice[OptimizerSweep, patch.OptimizerPatch],  # type: ignore
]




from py_gen_ml.cli_args.arg_refs import ArgRef
from py_gen_ml.cli_args.cli_func import pgml_cmd
from py_gen_ml.sweep.sweep import (
    BoolSweep,
    BytesSweep,
    Choice,
    FloatSweep,
    IntSweep,
    NestedChoice,
    StrSweep,
    Sweeper,
)
from py_gen_ml.sweep.tune.optuna import OptunaSampler
from py_gen_ml.yaml.yaml_model import YamlBaseModel

__all__ = [
    'YamlBaseModel',
    'Sweeper',
    'IntSweep',
    'FloatSweep',
    'BoolSweep',
    'StrSweep',
    'BytesSweep',
    'Choice',
    'NestedChoice',
    'ArgRef',
    'pgml_cmd',
    'OptunaSampler',
]



import os
from typing import Any

import torch
import torchvision
import torchvision.transforms as transforms


def get_transform() -> transforms.Compose:
    return transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ],)


def get_data_loader(transform: transforms.Compose, batch_size: int, train: bool) -> torch.utils.data.DataLoader[Any]:
    trainset = torchvision.datasets.CIFAR10(
        root=f"{os.environ['HOME']}/data/torchvision",
        train=train,
        download=train,
        transform=transform,
    )
    return torch.utils.data.DataLoader[Any](
        trainset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=4,
        prefetch_factor=4,
    )



# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml
import typing

import pydantic
import typer

from . import quickstart_b_base as base


class MLPQuickstartArgs(pgml.YamlBaseModel):
    """Multi-layer perceptron configuration"""

    num_layers: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of layers. Maps to 'num_layers'"),
        pydantic.Field(None),
        pgml.ArgRef("num_layers"),
    ]
    """Number of layers"""

    num_units: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of units. Maps to 'num_units'"),
        pydantic.Field(None),
        pgml.ArgRef("num_units"),
    ]
    """Number of units"""

    activation: typing.Annotated[
        typing.Optional[str],
        typer.Option(help="Activation function. Maps to 'activation'"),
        pydantic.Field(None),
        pgml.ArgRef("activation"),
    ]
    """Activation function"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import torch.nn


class Linear(pgml.YamlBaseModel):
    """Linear layer configuration"""

    in_features: int
    """Number of input features"""

    out_features: int
    """Number of output features"""

    bias: bool
    """Bias"""

    def build(self) -> "torch.nn.Linear":
        import torch.nn

        return torch.nn.Linear(
            in_features=self.in_features,
            out_features=self.out_features,
            bias=self.bias,
        )


class MLP(pgml.YamlBaseModel):
    """MLP configuration"""

    layers: typing.List[Linear]
    """Linear layers"""

    def build(self) -> "torch.nn.Sequential":
        import torch.nn

        return torch.nn.Sequential(
            *(elem.build() for elem in self.layers),
        )



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
ENABLE_FIELD_NUMBER: _ClassVar[int]
enable: _descriptor.FieldDescriptor
FACTORY_FIELD_NUMBER: _ClassVar[int]
factory: _descriptor.FieldDescriptor
CLI_FIELD_NUMBER: _ClassVar[int]
cli: _descriptor.FieldDescriptor
DEFAULT_FIELD_NUMBER: _ClassVar[int]
default: _descriptor.FieldDescriptor
AS_VARARGS_FIELD_NUMBER: _ClassVar[int]
as_varargs: _descriptor.FieldDescriptor

class Arg(_message.Message):
    __slots__ = ("name", "path")
    NAME_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    name: str
    path: str
    def __init__(self, name: _Optional[str] = ..., path: _Optional[str] = ...) -> None: ...

class CLI(_message.Message):
    __slots__ = ("arg", "enable")
    ARG_FIELD_NUMBER: _ClassVar[int]
    ENABLE_FIELD_NUMBER: _ClassVar[int]
    arg: _containers.RepeatedCompositeFieldContainer[Arg]
    enable: bool
    def __init__(self, arg: _Optional[_Iterable[_Union[Arg, _Mapping]]] = ..., enable: bool = ...) -> None: ...

class FieldDefaults(_message.Message):
    __slots__ = ("float", "double", "int32", "int64", "uint32", "uint64", "sint32", "sint64", "fixed32", "fixed64", "sfixed32", "sfixed64", "bool", "string", "bytes", "enum")
    FLOAT_FIELD_NUMBER: _ClassVar[int]
    DOUBLE_FIELD_NUMBER: _ClassVar[int]
    INT32_FIELD_NUMBER: _ClassVar[int]
    INT64_FIELD_NUMBER: _ClassVar[int]
    UINT32_FIELD_NUMBER: _ClassVar[int]
    UINT64_FIELD_NUMBER: _ClassVar[int]
    SINT32_FIELD_NUMBER: _ClassVar[int]
    SINT64_FIELD_NUMBER: _ClassVar[int]
    FIXED32_FIELD_NUMBER: _ClassVar[int]
    FIXED64_FIELD_NUMBER: _ClassVar[int]
    SFIXED32_FIELD_NUMBER: _ClassVar[int]
    SFIXED64_FIELD_NUMBER: _ClassVar[int]
    BOOL_FIELD_NUMBER: _ClassVar[int]
    STRING_FIELD_NUMBER: _ClassVar[int]
    BYTES_FIELD_NUMBER: _ClassVar[int]
    ENUM_FIELD_NUMBER: _ClassVar[int]
    float: float
    double: float
    int32: int
    int64: int
    uint32: int
    uint64: int
    sint32: int
    sint64: int
    fixed32: int
    fixed64: int
    sfixed32: int
    sfixed64: int
    bool: bool
    string: str
    bytes: bytes
    enum: str
    def __init__(self, float: _Optional[float] = ..., double: _Optional[float] = ..., int32: _Optional[int] = ..., int64: _Optional[int] = ..., uint32: _Optional[int] = ..., uint64: _Optional[int] = ..., sint32: _Optional[int] = ..., sint64: _Optional[int] = ..., fixed32: _Optional[int] = ..., fixed64: _Optional[int] = ..., sfixed32: _Optional[int] = ..., sfixed64: _Optional[int] = ..., bool: bool = ..., string: _Optional[str] = ..., bytes: _Optional[bytes] = ..., enum: _Optional[str] = ...) -> None: ...



# Autogenerated code. DO NOT EDIT.
import enum

import py_gen_ml as pgml


class Activation(str, enum.Enum):
    """Activation function"""

    RELU = "RELU"
    """ReLU activation"""

    GELU = "GELU"
    """Gelu activation"""


class MLP(pgml.YamlBaseModel):
    """MLP configuration"""

    activation: Activation
    """Activation function"""

    num_layers: int
    """Number of layers"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class LinearBlockPatch(pgml.YamlBaseModel):
    """Linear block configuration"""

    num_units: typing.Optional[int] = None
    """Number of units"""

    activation: typing.Optional[str] = None
    """Activation function"""


class OptimizerPatch(pgml.YamlBaseModel):
    """Optimizer configuration"""

    type: typing.Optional[str] = None
    """Type of optimizer"""

    learning_rate: typing.Optional[float] = None
    """Learning rate"""


class MLPPatch(pgml.YamlBaseModel):
    """Multi-layer perceptron configuration"""

    layers: typing.Optional[typing.List[LinearBlockPatch]] = None
    """List of linear blocks"""


class TrainingPatch(pgml.YamlBaseModel):
    """Training configuration"""

    mlp: typing.Optional[MLPPatch] = None
    """Multi-layer perceptron configuration"""

    optimizer: typing.Optional[OptimizerPatch] = None
    """Optimizer configuration"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml

from . import enum_demo_base as base


class MLPPatch(pgml.YamlBaseModel):
    """MLP configuration"""

    activation: typing.Optional[base.Activation] = None
    """Activation function"""

    num_layers: typing.Optional[int] = None
    """Number of layers"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml

from . import default_enum_base as base


class LinearPatch(pgml.YamlBaseModel):
    """Linear layer"""

    in_features: typing.Optional[int] = None
    """Number of input features"""

    out_features: typing.Optional[int] = None
    """Number of output features"""

    activation: typing.Optional[base.Activation] = None
    """Activation function"""



# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml
import typing

import pydantic
import typer

from . import dataloader_base as base


class DataLoaderConfigArgs(pgml.YamlBaseModel):
    """DataLoader configuration"""

    batch_size: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Batch size. Maps to 'batch_size'"),
        pydantic.Field(None),
        pgml.ArgRef("batch_size"),
    ]
    """Batch size"""

    num_workers: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of workers. Maps to 'num_workers'"),
        pydantic.Field(None),
        pgml.ArgRef("num_workers"),
    ]
    """Number of workers"""

    pin_memory: typing.Annotated[
        typing.Optional[bool],
        typer.Option(help="Pin memory. Maps to 'pin_memory'"),
        pydantic.Field(None),
        pgml.ArgRef("pin_memory"),
    ]
    """Pin memory"""

    persistent_workers: typing.Annotated[
        typing.Optional[bool],
        typer.Option(help="Persistent workers. Maps to 'persistent_workers'"),
        pydantic.Field(None),
        pgml.ArgRef("persistent_workers"),
    ]
    """Persistent workers"""

    prefetch_factor: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Prefetch factor. Maps to 'prefetch_factor'"),
        pydantic.Field(None),
        pgml.ArgRef("prefetch_factor"),
    ]
    """Prefetch factor"""
